from __future__ import annotations

__all__ = [
    'Chain', 'DualStageTransform', 'ImageTransform', 'MaskTransform',
    'Transform'
]

from collections.abc import Iterable
from typing import Any, Protocol, final, runtime_checkable

import numpy as np


@runtime_checkable
class Transform(Protocol):
    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        raise NotImplementedError

    def __mul__(self, prob: float) -> Transform:
        if not isinstance(prob, (int, float)):
            return NotImplemented
        if not (0 <= prob <= 1):
            raise ValueError('Probability should be in [0.0 .. 1.0] range')

        if prob == 1:
            return self

        if isinstance(self, _Maybe):
            prob *= self.prob
            self = self.func
        return _Maybe(self, prob)

    def __rmul__(self, prob: float) -> Transform:
        return self * prob

    def __or__(self, rhs: Transform) -> _OneOf:
        if not isinstance(rhs, Transform):
            return NotImplemented
        ts = (
            t_ for t in (self, rhs)
            for t_ in (t.transforms if isinstance(t, _OneOf) else [t]))
        return _OneOf(*ts)


class _SingleTransform(Transform):
    _key: str

    @final
    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        func = getattr(self, self._key)
        return {**data, self._key: func(data[self._key], rng)}


class ImageTransform(_SingleTransform):
    _key = 'image'

    def image(self, image: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        raise NotImplementedError


class MaskTransform(_SingleTransform):
    _key = 'mask'

    def mask(self, mask: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        raise NotImplementedError


class DualStageTransform(Transform):
    _keys = frozenset[str]({'image', 'mask'})

    def prepare(self, rng: np.random.Generator, /,
                **data) -> dict[str, Any] | None:
        return {}

    def image(self, image: np.ndarray, **params) -> np.ndarray:
        return image

    def mask(self, mask: np.ndarray, **params) -> np.ndarray:
        return mask

    @final
    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        if unknown_keys := {*data} - self._keys:
            raise ValueError(f'Got unknown keys in data: {unknown_keys}')

        params = self.prepare(rng, **data)
        if params is None:
            return data
        return {
            key: getattr(self, key)(value, **params)
            for key, value in data.items() if value is not None
        }


@final
class _Maybe(Transform):
    def __init__(self, func: Transform, prob: float = 1.0):
        self.func = func
        self.prob = prob

    def __repr__(self):
        return f'{self.prob:.2f} * {self.func!r}'

    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        return self.func(rng, **data) if rng.random() <= self.prob else data


class _Compose(Transform):
    transforms: tuple[Transform, ...]

    def _repr(self, words: Iterable[str]) -> str:
        if parts := ',\n'.join(words):
            parts = '\n'.join((f'    {p}' if p.strip() else p)
                              for p in parts.splitlines(True))
            parts = f'\n{parts}\n'
        return f'{type(self).__name__}({parts})'


@final
class Chain(_Compose):
    def __init__(self, *transforms: Transform):
        self.transforms = *(t for t in transforms
                            if not isinstance(t, _Maybe) or t.prob > 0),

    def __repr__(self) -> str:
        return self._repr(f'{t!r}' for t in self.transforms)

    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        for call in self.transforms:
            data = call(rng, **data)
        return data


@final
class _OneOf(_Compose):
    transforms: tuple[_Maybe, ...]
    probs: np.ndarray

    def __init__(self, *transforms: Transform):
        probs, self.transforms = zip(
            *((t.prob, t) if isinstance(t, _Maybe) else (1.0, _Maybe(t))
              for t in transforms))
        self.probs = np.array(probs)
        self.probs /= self.probs.sum()

    def __repr__(self) -> str:
        words = (f'{p:.2f} * {t.func}'
                 for p, t in zip(self.probs, self.transforms))
        return self._repr(words)

    def __call__(self, rng: np.random.Generator, /, **data) -> dict[str, Any]:
        func = rng.choice(self.transforms, p=self.probs)  # type: ignore
        return func.func(rng, **data)



from __future__ import annotations

__all__ = ['lock_seed', 'trace', 'trace_module', 'whereami']

import gc
import inspect
import os
import random
import types
from collections.abc import Iterator
from contextlib import suppress
from itertools import islice
from types import FrameType

import numpy as np
import wrapt

from ._import_hook import register_post_import_hook


def _get_module(frame: FrameType) -> str:
    if (module := inspect.getmodule(frame)) and module.__spec__:
        return module.__spec__.name
    return '__main__'


def _get_function(frame: FrameType) -> str:
    function = frame.f_code.co_name
    function = next(
        (f.__qualname__
         for f in gc.get_referrers(frame.f_code) if inspect.isfunction(f)),
        function)
    return '' if function == '<module>' else function


def _stack(frame: FrameType | None) -> Iterator[str]:
    while frame:
        yield f'{_get_module(frame)}:{_get_function(frame)}:{frame.f_lineno}'
        if frame.f_code.co_name == '<module>':  # Stop on module-level scope
            return
        frame = frame.f_back


def stack(skip: int = 0, limit: int | None = None) -> Iterator[str]:
    """Returns iterator of FrameInfos, stopping on module-level scope"""
    frame = inspect.currentframe()
    calls = _stack(frame)
    calls = islice(calls, skip + 1, None)  # Skip 'skip' outerless frames
    if not limit:
        return calls
    return islice(calls, limit)  # Keep at most `limit` outer frames


def whereami(skip: int = 0, limit: int | None = None) -> str:
    calls = stack(skip + 1, limit)
    return ' -> '.join(reversed([*calls]))


@wrapt.decorator
def trace(fn, _, args, kwargs):
    print(
        f'<({whereami(3)})> : {fn.__module__ or ""}.{fn.__qualname__}',
        flush=True)
    return fn(*args, **kwargs)


def _set_trace(obj, seen=None, prefix=None, module=None):
    # TODO: rewrite using unittest.mock
    if isinstance(obj, types.ModuleType):
        if seen is None:
            seen = set()
            prefix = obj.__name__
        if not obj.__name__.startswith(prefix) or obj.__name__ in seen:
            return
        seen.add(obj.__name__)
        for name in dir(obj):
            _set_trace(
                getattr(obj, name), module=obj, seen=seen, prefix=prefix)

    if not callable(obj):
        return

    if not hasattr(obj, '__dict__'):
        setattr(module, obj.__qualname__, trace(obj))
        print(f'wraps "{module.__name__}:{obj.__qualname__}"')
        return

    for name in obj.__dict__:
        with suppress(AttributeError, TypeError):
            member = getattr(obj, name)
            if not callable(member):
                continue
            decorated = trace(member)

            for m in (decorated, member, obj):
                with suppress(AttributeError):
                    decorated.__module__ = m.__module__
                    break
            else:  # noqa: PLW0120, RUF100
                decorated.__module__ = getattr(module, '__name__', '')
            setattr(obj, name, decorated)
            print(f'wraps "{module.__name__}:{obj.__qualname__}.{name}"')


def trace_module(name):
    """Enables call logging for each callable inside module name"""
    register_post_import_hook(_set_trace, name)


# ---------------------------------------------------------------------------


def lock_seed(seed: int) -> None:
    """Set seed for all modules: random/numpy/torch"""
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

    def _torch_seed(torch):
        import torch  # type: ignore[no-redef]
        import torch.backends.cudnn  # type: ignore[no-redef]

        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    register_post_import_hook(_torch_seed, 'torch')



"""
Fix for strange bug in SciPy on Anaconda for Windows
https://stackoverflow.com/questions/15457786/ctrl-c-crashes-python-after-importing-scipy-stats

Combines both:
.. [https://stackoverflow.com/a/39021051]
.. [https://stackoverflow.com/a/44822794]
"""
__all__ = ['apply']

import ctypes
import os
import sys
import warnings
from collections.abc import Iterator
from importlib.util import find_spec
from pathlib import Path

_FORTRAN_FLAG = 'FOR_DISABLE_CONSOLE_CTRL_HANDLER'


def _get_conda_libs() -> Iterator[ctypes.CDLL]:
    """Preload DLLs from icc_rt conda package"""
    root = Path(sys.prefix)
    if (root / 'conda-meta').exists() and find_spec('scipy'):
        for dllname in ('libmmd.dll', 'libifcoremd.dll'):
            if (dll := root / f'Library/bin/{dllname}').exists():
                yield ctypes.CDLL(dll.as_posix())


def _patch_handler_and_load_scipy() -> None:
    # Source - (stackoverflow)[https://stackoverflow.com/a/39021051/9868257]
    ptr = ctypes.c_void_p()
    ok = ctypes.windll.kernel32.VirtualProtect(
        ptr, ctypes.c_size_t(1), 0x40, ctypes.byref(ctypes.c_uint32(0)))
    if not ok or (addr := ptr.value) is None:
        return

    code: bytearray = (ctypes.c_char * 3).from_address(
        addr)  # type: ignore[assignment]
    if not code:
        return

    new = b'\xC2\x08\x00' if ctypes.sizeof(ctypes.c_void_p) == 4 else b'\xC3'
    patch_size = len(new)
    old, code[:patch_size] = code[:patch_size], new
    try:
        import scipy.stats  # noqa: F401
    finally:
        code[:patch_size] = old


def apply() -> None:
    if (sys.platform != 'win32'
            or _FORTRAN_FLAG in os.environ  # Patch is already applied
            or not [*_get_conda_libs()]):  # Only Anaconda's SciPy is affected
        return

    os.environ[_FORTRAN_FLAG] = '1'  # Child will inherit this, and work fine

    msg = ('Ctrl-C on Windows is broken when scipy is from conda. '
           'Please use scipy from PyPI')
    warnings.warn(msg, stacklevel=2)

    if 'scipy.stats' in sys.modules:
        msg2 = ('Cannot fix handling of Ctrl-C in current process. '
                'Import glow before scipy.stats to fix this.')
        warnings.warn(msg2, stacklevel=2)
    else:
        _patch_handler_and_load_scipy()



import os

import numpy as np

from glow import buffered


class Job:
    size = 20

    def __iter__(self):
        for i in range(self.size):
            m = np.random.rand(2048, 2048)
            (m @ m).sum()
            print(f'child[{os.getpid():5d}]: {i}')
            # if i == len(self) // 2:
            #     raise ValueError
            yield i

    def __len__(self):
        return self.size


if __name__ == '__main__':
    g = Job()
    # for i in buffered(g, mp=True):
    for i in buffered(g):
        if i == len(g) // 2:
            raise ValueError
        print(f'main [{os.getpid():5d}]: {i}')



from __future__ import annotations

__all__ = [
    'Cat', 'Encoder', 'Ensemble', 'Gate', 'Residual', 'ResidualCat', 'pre_norm'
]

from collections.abc import Iterable
from typing import Literal

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from .lazy import LazyLayerNorm
from .util import LazyNormFn


class Ensemble(nn.ModuleList):
    __constants__ = ['mode']

    def __init__(self, *branches: nn.Module | Iterable[nn.Module],
                 mode: Literal['sum', 'cat']):
        modules = (
            b if isinstance(b, nn.Module) else nn.Sequential(*b)
            for b in branches)
        super().__init__(modules)
        self.mode = mode

    def _sum(self, xs: list[torch.Tensor]) -> torch.Tensor:
        r = xs[0]
        for x in xs[1:]:
            r += x
        return r

    def _cat(self, xs: list[torch.Tensor]) -> torch.Tensor:
        return torch.cat(xs, dim=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        ys: list[torch.Tensor] = [m(x) for m in self]
        if self.mode == 'sum':
            return self._sum(ys)

        if self.mode == 'cat':
            return self._cat(ys)

        raise NotImplementedError


class _Sequential(nn.ModuleList):
    """
    This exists since TorchScript doesn't support inheritance, so the
    superclass method (_core_forward) needs to have a name other than `forward`
    that can be accessed in a subclass.
    See https://github.com/pytorch/pytorch/issues/42885.
    """
    def __init__(self, *modules: nn.Module) -> None:
        super().__init__(modules)

    def seq(self, x: torch.Tensor) -> torch.Tensor:
        # Replicates nn.Sequential.forward
        for m in self:
            x = m(x)
        return x


class ResidualCat(_Sequential):
    """Returns `cat([input, self(input)], dim=1)`, useful for U-Net"""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.cat([x, self.seq(x)], 1)


class Gate(_Sequential):
    """Returns input * self(input)"""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.seq(x)


class Residual(_Sequential):
    """Returns input + self(input)"""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.seq(x)


class Cat(_Sequential):
    """Computes the following:
    ```
    x = torch.cat(xs, 1)
    for m in modules:
        x = m(x)
    return x
    ```
    Checkpoints if makes sense.
    """
    def cat(self, xs: list[torch.Tensor]) -> torch.Tensor:
        x = torch.cat(xs, 1) if len(xs) != 1 else xs[0]
        return self.seq(x)

    @torch.jit.ignore
    def _cp_proxy(self, xs: list[torch.Tensor]) -> torch.Tensor:
        def closure(*xs: torch.Tensor) -> torch.Tensor:
            return self.cat([*xs])

        return checkpoint(closure, *xs)

    def forward(self, xs: list[torch.Tensor]) -> torch.Tensor:
        if torch.jit.is_scripting() or torch.jit.is_tracing():
            # checkpoint fails with JIT
            return self.cat(xs)

        if len(self) and any([x.requires_grad for x in xs]):  # noqa: C419
            return self._cp_proxy(xs)

        return self.cat(xs)


def pre_norm(*fn: nn.Module, norm: LazyNormFn = LazyLayerNorm) -> Residual:
    """Returns input + fn(norm(input))"""
    return Residual(norm(), *fn)


class Encoder(nn.Sequential):
    def __init__(self, stem: nn.Module, levels: nn.Module, head: nn.Module):
        super().__init__()
        self.stem = stem
        self.levels = levels
        self.head = head



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal, Optional, TypeVar

import pytest

from glow.cli import parse_args

T = TypeVar('T')


@dataclass
class Arg:
    arg: str


@dataclass
class List_:  # noqa: N801
    args: list[str]


@dataclass
class UntypedList:  # Forbidden, as list field should always be typed
    args: list


@dataclass
class UnsupportedSet:
    args: set[str]


@dataclass
class BadBoolean:  # Forbidden, as boolean field should always have default
    arg: bool


@dataclass
class Boolean:
    param: bool = False


@dataclass
class Nullable:
    param: Optional[str] = None  # noqa: UP007


@dataclass
class Optional_:  # noqa: N801
    param: str = 'hello'


@dataclass
class Nested:
    arg: str
    nested: Optional_


@dataclass
class NestedArg:  # Forbidden, as only top level args can be positional
    arg2: str
    nested: Arg


@dataclass
class Aliased:
    arg: str = 'hello'


@dataclass
class NestedAliased:  # Forbidden as all field names must be unique
    arg: str
    nested: Aliased


@pytest.mark.parametrize(('argv', 'expected'), [
    (['value'], Arg('value')),
    ([], List_([])),
    (['a'], List_(['a'])),
    (['a', 'b'], List_(['a', 'b'])),
    ([], Boolean()),
    (['--no-param'], Boolean()),
    (['--param'], Boolean(True)),
    ([], Nullable()),
    (['--param', 'value'], Nullable('value')),
    ([], Optional_()),
    (['--param', 'world'], Optional_('world')),
    (['value'], Nested('value', Optional_())),
    (['value', '--param', 'pvalue'], Nested('value', Optional_('pvalue'))),
])
def test_good_class(argv: list[str], expected: Any):
    cls = type(expected)
    result, _ = parse_args(cls, argv)
    assert isinstance(result, cls)
    assert result == expected


@pytest.mark.parametrize(('cls', 'exc_type'), [
    (Arg, SystemExit),
    (BadBoolean, ValueError),
    (UnsupportedSet, ValueError),
    (UntypedList, ValueError),
    (Nested, SystemExit),
    (NestedArg, ValueError),
    (NestedAliased, ValueError),
])
def test_bad_class(cls: type[Any], exc_type: type[BaseException]):
    with pytest.raises(exc_type):
        parse_args(cls, [])


def _no_op():
    return ()


def _arg(a: int):
    return a


def _kwarg(a: int = 4):
    return a


def _kwarg_opt(a: int = None):  # type: ignore[assignment]
    return a


def _kwarg_literal(a: Literal[1, 2] = 1):
    return a


def _kwarg_bool(a: bool = False):
    return a


def _kwarg_list(a: list[int] = []):  # noqa: B006
    return a


def _kwarg_opt_list(a: list[int] = None):  # type: ignore[assignment]
    return a


def _arg_kwarg(a: int, b: str = 'hello'):
    return a, b


@pytest.mark.parametrize(('argv', 'func', 'expected'), [
    ([], _no_op, ()),
    (['42'], _arg, 42),
    ([], _kwarg, 4),
    (['--a', '58'], _kwarg, 58),
    ([], _kwarg_opt, None),
    (['--a', '73'], _kwarg_opt, 73),
    ([], _kwarg_literal, 1),
    (['--a', '2'], _kwarg_literal, 2),
    ([], _kwarg_bool, False),
    (['--no-a'], _kwarg_bool, False),
    (['--a'], _kwarg_bool, True),
    ([], _kwarg_list, []),
    (['--a'], _kwarg_list, []),
    (['--a', '1'], _kwarg_list, [1]),
    (['--a', '1', '2'], _kwarg_list, [1, 2]),
    ([], _kwarg_opt_list, None),
    (['--a'], _kwarg_opt_list, []),
    (['--a', '1'], _kwarg_opt_list, [1]),
    (['--a', '1', '2'], _kwarg_opt_list, [1, 2]),
    (['53'], _arg_kwarg, (53, 'hello')),
    (['87', '--b', 'bye'], _arg_kwarg, (87, 'bye')),
])
def test_good_func(argv: list[str], func: Callable[..., T], expected: T):
    result, _ = parse_args(func, argv)
    assert result == expected



from collections import Counter
from uuid import UUID, uuid4

import pytest

from glow import Uid


def test_generation():
    assert 20 < len(str(Uid.v4())) < 24


def test_encoding():
    u = UUID('{3b1f8b40-222c-4a6e-b77e-779d5a94e21c}')
    assert str(Uid(u)) == 'CXc85b4rqinB7s5J52TRYb'


def test_decoding():
    u = UUID('{3b1f8b40-222c-4a6e-b77e-779d5a94e21c}')
    assert Uid('CXc85b4rqinB7s5J52TRYb') == u


def test_padding():
    uid = uuid4()
    uid_smallest = UUID(int=0)

    encoded = str(Uid(uid))
    assert Uid(encoded) == uid

    encoded_small = str(Uid(uid_smallest))
    assert Uid(encoded_small) == uid_smallest

    assert len(encoded) == len(encoded_small)


def test_stability():
    u = uuid4()
    assert u == Uid(str(Uid(u)))

    u = Uid.v4()
    assert isinstance(u, Uid)
    assert u == Uid(u)
    assert u == Uid(str(u))


def test_consistency():
    num_iterations = 1000
    lengths = Counter[int]()

    for _ in range(num_iterations):
        uid = Uid.v4()

        encoded = str(uid)
        lengths[len(encoded)] += 1
        uid_decoded = Uid(encoded)

        assert uid == uid_decoded

    assert len(lengths) == 1

    (_, count), = lengths.most_common()
    assert count == num_iterations


@pytest.mark.parametrize('x', [[], {}, ()])
def test_badtype(x):
    with pytest.raises(TypeError):
        Uid(x)


@pytest.mark.parametrize('x', ['', '0'])
def test_badvalue(x):
    with pytest.raises(ValueError):
        Uid(x)



from __future__ import annotations

__all__ = ['fc_densenet', 'max_vit', 'vit']

from collections.abc import Iterable
from itertools import accumulate
from typing import Literal

import torch
from einops.layers.torch import Rearrange, Reduce
from torch import nn

from .modules import (ConvCtx, DenseBlock, DenseDelta, Encoder, Ensemble,
                      LazyBias2d, MaxVitBlock, VitBlock)


def fc_densenet(num_classes: int,
                init: int = 48,
                depths: Iterable[int] = (4, 4),
                step: int = 12,
                bottleneck: bool = False):
    # TODO: define Fold/Unfold or Stash/Braid module
    # TODO: use flat structure instead of recursive nesting used currently
    """
    Implementation of
    [The One Hundred Layers Tiramisu](https://arxiv.org/pdf/1611.09326.pdf)

    See [FC-DenseNet](https://github.com/SimJeg/FC-DenseNet) as reference
    Theano/Lasagne-based implementation,
    and (https://github.com/bfortuner/pytorch_tiramisu) as pytorch fork.
    """
    *depths, = depths
    *dims, _ = accumulate([init] + [step * depth for depth in depths])

    core: list[nn.Module] = []

    ctx = ConvCtx(parity=0)
    depth = depths.pop()
    while depths:
        dim = dims.pop()
        branch = [
            ctx.norm_act_conv(ctx.conv(dim, 1)),
            ctx.max_pool(),
            *core,
            DenseDelta(depth, step, bottleneck, ctx),
            # ? norm + act
            ctx.conv_unpool(depth * step, bias=False),
        ]
        depth = depths.pop()
        core = [
            DenseBlock(depth, step, bottleneck, ctx),
            Ensemble(nn.Identity(), branch, mode='cat'),
        ]

    core = [
        ctx.conv(init, 3, bias=False),
        *core,
        DenseBlock(depth, step, bottleneck, ctx),
        # ? norm + act
        ctx.conv(num_classes, 1),
    ]
    return nn.Sequential(*core)


# ------------------------------ token actions -------------------------------


class CatToken(nn.Module):  # [B, N, D] -> [B, 1 + N, D]
    def __init__(self,
                 dim: int,
                 count: int = 1,
                 device: torch.device | None = None):
        super().__init__()
        self.token = nn.Parameter(torch.empty(count, dim, device=device))
        nn.init.normal_(self.token)

    def extra_repr(self) -> str:
        num_tokens, features = self.token.shape
        return f'{num_tokens=}, {features=}'

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b = x.shape[0]
        tokens = self.token.broadcast_to(b, -1, -1)
        return torch.cat((tokens, x), dim=1)


class PopToken(nn.Module):  # [B, N, D] -> [B, D]
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x[:, 0, :]


# -------------------------------- models ------------------------------------


def vit(num_classes: int,
        patch_size: int,
        depth: int,
        dim: int,
        dim_head: int = 64,
        mlp_ratio: float = 4.,
        pool: Literal['cls', 'mean'] = 'cls',
        dropout: float = 0.,
        dropout_emb: float = 0.,
        qkv_bias: bool = True,
        reattn: bool = False) -> Encoder:
    assert pool in {'cls', 'mean'}

    stem = nn.Sequential(
        nn.LazyConv2d(dim, patch_size, patch_size),
        LazyBias2d(),
        Rearrange('b c h w -> b (h w) c'),
    )

    # ViT (Vision Transformer)
    #   = patchify (convAxA/sA) + cat-token + pop-token
    # DeepViT
    #   = ViT + reattention
    # CVT (Compact Vision Transformer)
    #   = patchify + identity + mean
    # CCT (Compact Convolutional Transformer)
    #   = CVT + patchify replaced with one of
    #   - 1-2 x [conv3x3 - relu - maxpool(2,2)], i.e. eff.stride = 2-4
    #   - 1-2 x [conv7x7/s2 - relu - maxpool(2,2)], i.e. eff.stride = 4-16
    transformer = nn.Sequential(
        # Operates in (b n d) domain
        nn.Identity() if pool == 'mean' else CatToken(dim),
        nn.Dropout(dropout_emb, inplace=True),
        *(VitBlock(dim, dim_head, mlp_ratio, dropout, qkv_bias, reattn)
          for _ in range(depth)),
        Reduce('b n d -> b d', 'mean') if pool == 'mean' else PopToken(),
    )
    head = nn.Sequential(
        nn.LayerNorm(dim),
        nn.Linear(dim, num_classes),
    )
    return Encoder(stem, transformer, head)


def max_vit(num_classes: int,
            depths: tuple[int, ...],
            dim: int,
            dim_head: int = 32,
            dim_stem: int | None = None,
            window_size: int = 7,
            bn_ratio: float = 4,
            se_ratio: float = 0.25,
            mlp_ratio: float = 4,
            qkv_bias: bool = False,
            dropout: float = 0.1) -> Encoder:
    dim_stem = dim_stem or dim
    # ctx = ConvCtx(activation=nn.GELU, inplace=False, overlap=0)  # 1
    # ctx = ConvCtx(activation=nn.GELU, inplace=False, parity=0, overlap=0)  # 2
    # ctx = ConvCtx(activation=nn.GELU, inplace=False)  # 3
    ctx = ConvCtx(activation=nn.GELU, inplace=False, parity=0)  # 4
    # ctx = ConvCtx(activation=nn.SiLU, parity=0)

    stem = nn.Sequential(
        ctx.conv(dim_stem, stride=2, bias=False),
        ctx.conv(dim_stem, 3),
    )

    dims = *(dim << i for i, _ in enumerate(depths)),
    dims = (dim_stem, *dims)

    layers: list[nn.Module] = []
    for dim, depth in zip(dims[1:], depths):
        layers += [
            MaxVitBlock(dim, dim_head, window_size, stride, bn_ratio, se_ratio,
                        mlp_ratio, dropout, qkv_bias, ctx)
            for stride in [2] + [1] * (depth - 1)
        ]

    head = nn.Sequential(
        Reduce('b d h w -> b d', 'mean'),
        nn.LayerNorm(dims[-1]),
        nn.Linear(dims[-1], num_classes),
    )
    return Encoder(stem, nn.Sequential(*layers), head)



"""
Patches builtin `print` function to be compatible with `tqdm`.
Adds some thread safety.
"""
__all__ = ['apply']

import builtins
from functools import update_wrapper, wraps
from threading import RLock

from ._import_hook import register_post_import_hook

_print = builtins.print
_lock = RLock()


@wraps(_print)
def locked_print(*args, **kwargs):
    with _lock:
        _print(*args, **kwargs)


def tqdm_print(module, *values, sep=' ', end='\n', file=None, flush=False):
    module.tqdm.write(sep.join(map(str, values)), end=end, file=file)


def patch_print(module) -> None:
    # Create blank to force initialization of cls._lock and cls._instances
    tqdm = module.tqdm
    tqdm(disable=True)

    def tqdm_print(*values, sep=' ', end='\n', file=None, flush=False):
        tqdm.write(sep.join(map(str, values)), end=end, file=file)

    builtins.print = update_wrapper(tqdm_print, _print)


def apply() -> None:
    builtins.print = locked_print
    register_post_import_hook(patch_print, 'tqdm')



import time
import weakref
from multiprocessing.shared_memory import SharedMemory

import numpy as np
from loky import get_reusable_executor

# import sys
# from loky.backend import resource_tracker
# sys.modules['multiprocessing.resource_tracker'] = resource_tracker


class SharedArray:
    def __init__(self, arr):
        self.shm = SharedMemory(create=True, size=arr.nbytes)
        self.shm.buf[:] = arr.view(np.uint8)
        self.shape = arr.shape
        self.dtype = arr.dtype
        weakref.finalize(self, self.shm.unlink)

    def rebuild(self):
        return np.asarray(self.shm.buf).view(self.dtype).reshape(self.shape)

    def mean(self):
        time.sleep(2)
        return self.rebuild().mean()


if __name__ == '__main__':
    # executor = get_reusable_executor(max_workers=1, context='loky_init_main')
    executor = get_reusable_executor(max_workers=1)

    # Create 1GB shared array
    arr = SharedArray(np.random.rand(int(1.25e8)))
    mean = arr.mean()

    f = executor.submit(arr.mean)
    assert f.result() == mean



import argparse
import pathlib
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import RandomSampler
from torchvision import transforms as tfs
from torchvision.datasets import CIFAR10
from tqdm.auto import tqdm

import glow
import glow.metrics as m
import glow.nn as gnn

DEVICE = torch.device('cuda')

glow.lock_seed(42)
torch.backends.cudnn.benchmark = True  # type: ignore
rg = np.random.default_rng()

# ------------------------------ define model ------------------------------


def make_model_default():
    return nn.Sequential(
        nn.Conv2d(3, 6, 5),  # > 28^2
        nn.ReLU(),
        nn.MaxPool2d(2),  # > 14^2
        nn.Conv2d(6, 16, 5),  # > 10^2
        nn.ReLU(),
        nn.MaxPool2d(2),  # > 5^2
        gnn.View(-1),  # > 1:400
        nn.Linear(400, 120),
        nn.ReLU(),
        nn.Linear(120, 84),
        nn.ReLU(),
        nn.Linear(84, 10),
    )


def make_model_new(init=16):
    def conv(cin, cout=None, groups=1, pad=2, stride=1):
        cout = cout or cin
        ksize = stride + pad * 2
        return nn.Sequential(
            nn.Conv2d(
                cin, cout, ksize, stride, pad, groups=groups, bias=False),
            nn.BatchNorm2d(cout),
            nn.ReLU(inplace=True),
        )

    def conv_down(cin, cout=None):
        cout = cout or cin
        return nn.Sequential(
            conv(cin, cout, pad=1, stride=2),
            gnn.Sum(
                conv(cout, cout * 2),
                conv(cout * 2, pad=2, groups=cout * 2),
                conv(cout * 2, cout)[:-1],
                tail=nn.ReLU(),
                skip=0.1),
        )

    return nn.Sequential(
        conv_down(3, init),  # > 16^2
        conv_down(init, init * 2),  # > 8^2
        conv_down(init * 2, init * 4),  # > 4^2
        conv(init * 4, init * 8, pad=2),
        nn.AdaptiveAvgPool2d(1),  # > 1
        gnn.View(-1),
        nn.Linear(init * 8, 10),
    )


# parse args

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument(
    'root', type=pathlib.Path, help='location of cifar10/ folder')
parser.add_argument('--batch-size', type=int, default=4, help='train batch')
parser.add_argument('--epochs', type=int, default=12, help='count of epochs')
parser.add_argument('--steps-per-epoch', type=int, help='steps per epoch')
parser.add_argument('--width', type=int, default=32, help='width of network')
parser.add_argument(
    '--fp16', action='store_true', help='enable mixed precision mode')
parser.add_argument('--plot', action='store_true', help='enable plot')

args = parser.parse_args()

epoch_len = (8000 // args.batch_size
             if args.steps_per_epoch is None else args.steps_per_epoch)
sample_size = args.epochs * epoch_len * args.batch_size

tft = tfs.Compose([
    tfs.RandomCrop(32, padding=4),
    tfs.RandomHorizontalFlip(),
    tfs.ToTensor(),
])
ds = CIFAR10(args.root / 'cifar10', transform=tft, download=True)
ds_val = CIFAR10(args.root / 'cifar10', transform=tfs.ToTensor(), train=False)

loader = gnn.make_loader(
    ds,
    args.batch_size,
    sampler=RandomSampler(ds, True, sample_size),
    multiprocessing=False)
val_loader = gnn.make_loader(ds_val, 100, multiprocessing=False)

# net = make_model_default()
net = make_model_new(args.width)
net.to(DEVICE)
opt = torch.optim.AdamW(net.parameters())
print(gnn.param_count(net))

criterion = nn.CrossEntropyLoss()
metrics = [
    m.Lambda(criterion, name='loss'),
    m.Confusion(acc=m.accuracy, kappa=m.kappa),
]
stepper = gnn.Stepper(
    net, opt, criterion, metrics, device=DEVICE, fp16=args.fp16)

history = defaultdict[str, list](list)
with tqdm(total=epoch_len * args.epochs, desc='train') as pbar:
    for i, split in enumerate(glow.ichunked(loader, epoch_len), 1):
        tscalars = stepper.run(split, pbar).scalars
        with tqdm(total=len(val_loader), desc='val', leave=False) as pbar_val:
            vscalars = stepper.run(val_loader, pbar_val, False).scalars

        _tags = sorted({*tscalars, *vscalars})
        scores = {tag: [tscalars[tag], vscalars[tag]] for tag in _tags}

        for tag, values in scores.items():
            history[tag].append(values)
        scalars_fmt = ', '.join(
            f'{t}: {{:.3f}}/{{:.3f}}'.format(*vs) for t, vs in scores.items())
        print(f'[{i:03d}] {scalars_fmt}')

if args.plot:
    fig = plt.figure(f'batch_size={args.batch_size}, fp16={args.fp16}')
    for i, (tag, values_) in enumerate(history.items(), 1):
        ax = fig.add_subplot(1, len(history), i)
        ax.legend(ax.plot(values_), ['train', 'val'])
        ax.set_title(tag)
        ax.set_ylim([
            int(min(x for xs in values_ for x in xs)),
            int(max(x for xs in values_ for x in xs) + 0.999)
        ])
    plt.show()



__all__ = ['Show']

import weakref

import cv2
import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn

# TODO: rewrite like `def traced(nn.Module) -> nn.Module`
# TODO: use pyqt/matplotlib to create window


class Show(nn.Module):
    """Shows contents of tensors during forward pass"""
    nsigmas = 2
    weight: torch.Tensor
    bias: torch.Tensor

    def __init__(self, colored: bool = False):
        super().__init__()
        self.name = f'{type(self).__name__}_0x{id(self):x}'
        self.colored = colored

        weight = torch.tensor(128 / self.nsigmas)
        bias = torch.tensor(128.)
        self.register_buffer('weight', weight, persistent=False)
        self.register_buffer('bias', bias, persistent=False)

        self.close = weakref.finalize(self, cv2.destroyWindow, self.name)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        c = inputs.shape[1]

        with torch.no_grad():
            weight = self.weight.expand(c)
            bias = self.bias.expand(c)
            ten = F.instance_norm(inputs, None, None, weight, bias)
            image: np.ndarray = ten.clamp_(0, 255).byte().cpu().numpy()

        if self.colored:
            groups = c // 3
            image = image[:, :groups * 3, :, :]
            image = rearrange(image, 'b (g c) h w -> (b h) (g w) c', c=3)
        else:
            image = rearrange(image, 'b c h w -> (b h) (c w)')

        cv2.imshow(self.name, image)
        cv2.waitKey(1)

        return inputs



__all__ = ['dither']

from types import MappingProxyType
from typing import Literal

import cv2
import numba
import numpy as np

_MATRICES = MappingProxyType({
    key: cv2.normalize(np.array(mat, 'f4'), None, norm_type=cv2.NORM_L1)
    for key, mat in {
        'jarvis-judice-ninke': [
            [0, 0, 0, 7, 5],
            [3, 5, 7, 5, 3],
            [1, 3, 5, 3, 1],
        ],
        'sierra': [
            [0, 0, 0, 5, 3],
            [2, 4, 5, 4, 2],
            [0, 2, 3, 2, 0],
        ],
        'stucki': [
            [0, 0, 0, 8, 4],
            [2, 4, 8, 4, 2],
            [1, 2, 4, 2, 1],
        ],
    }.items()
})
_DitherKind = Literal['jarvis-judice-ninke', 'stucki', 'sierra']


@numba.jit
def _dither(image: np.ndarray, mat: np.ndarray, quant: int) -> np.ndarray:
    for y in range(image.shape[0] - 2):
        for x in range(2, image.shape[1] - 2):
            old = image[y, x]
            new = (old // quant) * quant
            delta = ((old - new) * mat).astype(image.dtype)
            image[y:y + 3, x - 2:x + 3] += delta
            image[y, x] = new

    return image[:-2, 2:-2]


def dither(image: np.ndarray,
           bits: int = 3,
           kind: _DitherKind = 'stucki') -> np.ndarray:
    mat = _MATRICES[kind]
    assert image.dtype == 'u1'
    if image.ndim == 3:
        mat = mat[..., None]
        channel_pad = [[0, 0]]
    else:
        channel_pad = []
    image = np.pad(image, [[0, 2], [2, 2], *channel_pad])

    dtype = image.dtype
    if dtype == 'uint8':
        max_value = 256
        image = image.astype('i2')
    else:
        max_value = 1
    quant = max_value / 2 ** bits

    image = _dither(image, mat, quant)
    return image.clip(0, max_value - quant).astype(dtype)



__all__ = [
    'Confusion', 'ConfusionGrad', 'accuracy', 'accuracy_balanced', 'iou',
    'kappa', 'kappa_quadratic_weighted'
]

import torch
import torch.distributed as dist
from torch.distributed import nn

from .base import Staged, to_index_sparse, to_prob_sparse

_EPS = torch.finfo(torch.float).eps


class Confusion(Staged):
    """Confusion Matrix. Returns 2d tensor"""
    def __call__(self, pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
        c, pred, true = to_index_sparse(pred, true)

        mat = torch.zeros(c, c, dtype=torch.long)
        mat.index_put_((true, pred), torch.tensor(1), accumulate=True)

        return mat.double() / mat.sum()

    def collect(self, mat: torch.Tensor) -> dict[str, torch.Tensor]:
        c = mat.shape[0]
        return {f'cm{c}': mat, **super().collect(mat)}


class ConfusionGrad(Confusion):
    """Confusion Matrix which can be used for loss functions"""
    def __call__(self, pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
        c, pred, true = to_prob_sparse(pred, true)

        assert pred.dtype == torch.float32
        mat = pred.new_zeros(c, c).index_add(0, true, pred)

        if dist.is_initialized() and dist.get_world_size() > 1:
            mat = nn.all_reduce(mat)
        return mat.double() / mat.sum()


def accuracy(mat: torch.Tensor) -> torch.Tensor:
    return mat.trace() / mat.sum().clamp(_EPS)


def accuracy_balanced(mat: torch.Tensor) -> torch.Tensor:
    return (mat.diag() / mat.sum(1).clamp(_EPS)).mean()


def kappa(mat: torch.Tensor) -> torch.Tensor:
    expected = mat.sum(0) @ mat.sum(1)
    observed = mat.trace()
    return 1 - (1 - observed) / (1 - expected).clamp(_EPS)


def kappa_quadratic_weighted(mat: torch.Tensor) -> torch.Tensor:
    y, x = map(torch.arange, mat.shape)

    weights = (y[:, None] - x[None, :]).double() ** 2
    weights /= weights.max()

    expected = mat.sum(0) @ weights @ mat.sum(1)
    observed = mat.view(-1) @ weights.view(-1)
    return 1 - observed / expected.clamp(_EPS)


def iou(mat: torch.Tensor) -> torch.Tensor:
    return mat.diag() / (mat.sum(0) + mat.sum(1) - mat.diag()).clamp(_EPS)


def dice(mat: torch.Tensor) -> torch.Tensor:
    return 2 * mat.diag() / (mat.sum(0) + mat.sum(1)).clamp(_EPS)



from __future__ import annotations

__all__ = [
    'call_once', 'shared_call', 'streaming', 'threadlocal', 'weak_memoize'
]

import sys
import threading
from collections.abc import Callable, Hashable, Iterable, Sequence
from concurrent.futures import Future, wait
from dataclasses import dataclass, field
from functools import partial, update_wrapper
from queue import Empty, SimpleQueue
from threading import Lock, Thread
from time import monotonic, sleep
from typing import Any, TypeVar, cast
from weakref import WeakValueDictionary

from .util import make_key

_T = TypeVar('_T')
_R = TypeVar('_R')
_F = TypeVar('_F', bound=Callable)
_BatchFn = Callable[[list[_T]], Iterable[_R]]
_ZeroArgsF = TypeVar('_ZeroArgsF', bound=Callable[[], object])

_PATIENCE = 0.01
_unset = object()


def threadlocal(fn: Callable[..., _T], *args: object,
                **kwargs: object) -> Callable[[], _T]:
    """Thread-local singleton factory, mimics `functools.partial`"""
    local_ = threading.local()

    def wrapper() -> _T:
        try:
            return local_.obj
        except AttributeError:
            local_.obj = fn(*args, **kwargs)
            return local_.obj

    return update_wrapper(wrapper, fn)


@dataclass
class _UFuture:
    _fn: Callable[[], object]
    _lock: Lock = field(default_factory=Lock)
    _result: object = _unset
    _exception: BaseException | None = None

    def result(self):
        with self._lock:
            if self._exception:
                raise self._exception
            if self._result is not _unset:
                return self._result

            try:
                self._result = r = self._fn()
            except BaseException as e:
                self._exception = e
                raise
            return r


def call_once(fn: _ZeroArgsF) -> _ZeroArgsF:
    """Makes callable a singleton.

    DO NOT USE with recursive functions"""
    def wrapper():
        return uf.result()

    fn._future = uf = _UFuture(fn)  # type: ignore[attr-defined]
    return cast(_ZeroArgsF, update_wrapper(wrapper, fn))


def shared_call(fn: _F, /) -> _F:
    """Merges duplicate parallel invocations of callable to a single one.

    DO NOT USE with recursive functions"""
    fs = WeakValueDictionary[Hashable, _UFuture]()
    lock = Lock()

    def wrapper(*args, **kwargs):
        key = make_key(*args, **kwargs)

        with lock:  # Create only one task per args-kwargs set
            if not (uf := fs.get(key)):
                fs[key] = uf = _UFuture(partial(fn, *args, **kwargs))

        return uf.result()

    return cast(_F, update_wrapper(wrapper, fn))


def weak_memoize(fn: _F, /) -> _F:
    """Preserves each result of each call until they are garbage collected."""
    rs = WeakValueDictionary[Hashable, Any]()
    lock = Lock()

    def wrapper(*args, **kwargs):
        key = make_key(*args, **kwargs)

        with lock:
            if (r := rs.get(key, _unset)) is not _unset:
                return r

        r = fn(*args, **kwargs)

        with lock:
            rs[key] = r
        return r

    return cast(_F, update_wrapper(wrapper, fn))


# ----------------------------- batch collation ------------------------------


def _fetch_batch(q: SimpleQueue[_T], batch_size: int,
                 timeout: float) -> list[_T]:
    batch: list[_T] = []

    # Wait indefinitely until the first item is received
    if sys.platform == 'win32':
        # On Windows lock.acquire called without a timeout is not interruptible
        # See:
        # https://bugs.python.org/issue29971
        # https://github.com/dask/dask/pull/2144#issuecomment-290556996
        # https://github.com/dask/dask/pull/2144/files
        while not batch:
            try:
                batch.append(q.get(timeout=_PATIENCE))
            except Empty:
                sleep(0)  # Allow other thread to fill the batch
    else:
        batch.append(q.get())

    endtime = monotonic() + timeout
    while len(batch) < batch_size and (waittime := endtime - monotonic()) > 0:
        try:
            batch.append(q.get(timeout=waittime))
        except Empty:
            break
    return batch


def _batch_invoke(
    func: _BatchFn[_T, _R],
    batch: Sequence[tuple[Future[_R], _T]],
):
    batch = [(f, x) for f, x in batch if f.set_running_or_notify_cancel()]
    if not batch:
        return

    try:
        *results, = func([x for _, x in batch])
        if len(results) != len(batch):
            raise RuntimeError(  # noqa: TRY301
                'Input batch size is not equal to output: '
                f'{len(results)} != {len(batch)}')

    except BaseException as exc:  # noqa: BLE001
        for f, _ in batch:
            f.set_exception(exc)

    else:
        # TODO: use zip(strict=True) for python3.10+ when 3.9 is EOL
        for (f, _), r in zip(batch, results):
            f.set_result(r)


def _start_fetch_compute(func, workers, batch_size, timeout):
    q = SimpleQueue()  # type: ignore[var-annotated]
    lock = Lock()

    def loop():
        while True:
            # Because of lock, _fetch_batch could be inlined into wrapper,
            # and dispatch to thread pool could be done from there,
            # thus allowing usage of scalable ThreadPool
            # TODO: implement above
            with lock:  # Ensurance that none worker steals tasks from other
                batch = _fetch_batch(q, batch_size, timeout)
            if batch:
                _batch_invoke(func, batch)
            else:
                sleep(0.001)

    for _ in range(workers):
        Thread(target=loop, daemon=True).start()
    return q


def streaming(func=None,
              /,
              *,
              batch_size,
              timeout=0.1,
              workers=1,
              pool_timeout=20.):
    """
    Delays start of computation to until batch is collected.
    Accepts two timeouts (in seconds):
    - `timeout` is a time to wait till the batch is full, i.e. latency.
    - `pool_timeout` is time to wait for results.

    Uses ideas from
    - https://github.com/ShannonAI/service-streamer
    - https://github.com/leon0707/batch_processor
    - ray.serve.batch
      https://github.com/ray-project/ray/blob/master/python/ray/serve/batching.py

    Note: currently supports only functions and bound methods.

    Implementation details:
    - constantly keeps alive N workers
    - any caller enqueues jobs and starts waiting
    - on any failure during waiting caller cancels all jobs it submitted
    - single worker at a time fetches jobs from shared queue, resolves them,
      and notifies all waiters
    """
    if func is None:
        return partial(
            streaming,
            batch_size=batch_size,
            timeout=timeout,
            workers=workers,
            pool_timeout=pool_timeout)

    assert callable(func)
    assert workers >= 1
    q = _start_fetch_compute(func, workers, batch_size, timeout)

    def wrapper(items):
        fs = {Future(): item for item in items}  # type: ignore[var-annotated]
        try:
            for f_x in fs.items():
                q.put(f_x)  # Schedule task
            dnd = wait(fs, pool_timeout, return_when='FIRST_EXCEPTION')

        finally:  # Cancel all not-yet-running tasks, we're beyond deadline
            for f in fs:
                f.cancel()

        if dnd.not_done:  # Some tasks timed out
            del dnd, fs  # ? Break reference cycle
            raise TimeoutError
        return [f.result() for f in fs]  # Cannot time out - all are done

    # TODO: if func is instance method - recreate wrapper per instance
    # TODO: find how to distinguish between
    # TODO:  not yet bound method and plain function
    # TODO:  maybe implement __get__ on wrapper
    return update_wrapper(wrapper, func)



from .aggregates import (Encoder, Ensemble, Gate, Residual, ResidualCat,
                         pre_norm)
from .context import ConvCtx
from .convnets import (BottleneckResidualBlock, DenseBlock, DenseDelta,
                       ResNeXtBlock, SplitAttention, SqueezeExcitation, mbconv,
                       mobilenet_v2_block, mobilenet_v3_block, resnest_block)
from .lazy import LazyBias2d, LazyConv2dWs, LazyGroupNorm, LazyLayerNorm
from .simple import Decimate2d, Noise, Upscale2d
from .transformers import (Attention, FeedForward, MaxVitBlock,
                           MultiAxisAttention, VitBlock)
from .vision import Show

__all__ = [
    'Attention',
    'BottleneckResidualBlock',
    'ConvCtx',
    'Decimate2d',
    'DenseBlock',
    'DenseDelta',
    'Encoder',
    'Ensemble',
    'FeedForward',
    'Gate',
    'LazyBias2d',
    'LazyConv2dWs',
    'LazyGroupNorm',
    'LazyLayerNorm',
    'MaxVitBlock',
    'MultiAxisAttention',
    'Noise',
    'ResNeXtBlock',
    'Residual',
    'ResidualCat',
    'Show',
    'SplitAttention',
    'SqueezeExcitation',
    'Upscale2d',
    'VitBlock',
    'mbconv',
    'mobilenet_v2_block',
    'mobilenet_v3_block',
    'pre_norm',
    'resnest_block',
]



from dataclasses import dataclass

import numpy as np
import pytest

import glow
from glow import eat
from glow.core._parallel import max_cpu_count

DEATH_RATE = 0
SIZE = 100
NUM_STEPS = 10
DTYPE = np.dtype(np.float32)

MAX_PROCS = 8
NUM_PROCS = max_cpu_count(MAX_PROCS, mp=True)


def _make_array(n):
    return np.arange(int(n), dtype=DTYPE)


class AsInit:
    def __init__(self, n):
        self.data = _make_array(n)

    def args(self):
        return [[np.mean(self.data)]] * NUM_STEPS

    def __call__(self, mean):
        assert np.mean(self.data) == mean


@dataclass(frozen=True)
class AsArg:
    n: int

    def args(self):
        for _ in range(NUM_STEPS):
            data = _make_array(self.n)
            yield data, np.mean(data)

    def __call__(self, data, mean):
        return np.mean(data) == mean


class AsArgRepeated(AsArg):
    def args(self):
        args, = super().args()
        return [[*args]] * NUM_STEPS


@dataclass(frozen=True)
class AsResult:
    n: int

    def args(self):
        return np.arange(NUM_STEPS).reshape(-1, 1)

    def __call__(self, _):
        return _make_array(self.n)


def run_glow(task, *args):
    return glow.map_n(task, *args, max_workers=2, mp=True)


def run_joblib(task, *args):
    from glow.joblib import Parallel, delayed
    return Parallel(
        n_jobs=2, backend='multiprocessing')(
            delayed(task)(*a) for a in zip(*args))


def run_joblib_mp(task, *args):
    from glow.joblib import Parallel, delayed
    return Parallel(n_jobs=2)(delayed(task)(*a) for a in zip(*args))


def bench_ipc_speed(order=25, steps=100):
    from matplotlib import pyplot as plt

    sizes = np.asarray([2 ** order, *np.logspace(order, 2, num=steps, base=2)])
    to_bytes = DTYPE.itemsize * 2  # x2, because copy+read

    fig = plt.figure(figsize=(10, 4))
    workers = (AsInit, AsArgRepeated, AsArg, AsResult)
    for i, worker in enumerate(workers, 1):
        ax = fig.add_subplot(
            1,
            len(workers),
            i,
            ylabel='bytes/s',
            xlabel='size',
            xscale='log',
            yscale='log',
            ylim=(1, 1e12),
            title=worker.__name__)
        for runner in [run_glow, run_joblib, run_joblib_mp]:
            label = f'{worker.__name__}-{runner.__name__}'
            times: list[int] = []
            for size in sizes:
                task = worker(size)
                args = zip(*task.args())  # type: ignore[attr-defined]
                with glow.timer(times.append):
                    eat(runner(task, *args))

            bps = to_bytes * NUM_STEPS * sizes / np.asarray(times)
            print(f'max {glow.si_bin(bps.max())}/s - {label}')

            ax.plot(to_bytes * sizes, bps, label=runner.__name__)
            ax.legend()

    plt.tight_layout()
    plt.show()


def source(size):
    deads = np.random.uniform(size=size).astype('f4')
    print(np.where(deads < DEATH_RATE)[0].tolist()[:10])
    for seed, death in enumerate(deads):
        if death < DEATH_RATE:
            raise ValueError(f'Source died: {seed}')
        yield seed


def do_work(seed, offset):
    rg = np.random.default_rng(seed + offset)
    n = 10
    a = rg.random((2 ** n, 2 ** n), dtype='f4')
    b = rg.random((2 ** n, 2 ** n), dtype='f4')
    (a @ b).sum()
    if rg.uniform() < DEATH_RATE:
        raise ValueError(f'Worker died: {seed}') from None
    return seed


def _test_interrupt():
    """Should die gracefully on Ctrl-C"""
    sources = (
        source(SIZE),
        np.random.randint(2 ** 10, size=SIZE),
    )
    # sources = map(glow.buffered, sources)
    res = glow.map_n(do_work, *sources, max_workers=NUM_PROCS, mp=True)
    print('start main', end='')
    for r in res:
        print(end=f'\rmain {r} computes...')
        rg = np.random.default_rng(r)
        n = 10
        a = rg.random((2 ** n, 2 ** n), dtype='f4')
        b = rg.random((2 ** n, 2 ** n), dtype='f4')
        (a @ b).sum()
        yield r
        print(end=f'\rmain {r} waits...')
    print('\rmain done')


@pytest.mark.skipif(NUM_PROCS < 2, reason='not enough memory')
def test_interrupt():
    rs = _test_interrupt()
    assert [*rs] == [*range(SIZE)]


@pytest.mark.skipif(NUM_PROCS < 2, reason='not enough memory')
def test_interrupt_with_buffer():
    rs = glow.buffered(_test_interrupt())
    assert [*rs] == [*range(SIZE)]


if __name__ == '__main__':
    bench_ipc_speed()



from __future__ import annotations

__all__ = ['DdpSampler']

from collections.abc import Iterator, Sized
from itertools import chain, cycle, islice
from typing import Generic, Protocol, TypeVar

import torch
from torch.utils.data import (RandomSampler, Sampler, SequentialSampler,
                              SubsetRandomSampler, WeightedRandomSampler)

from ..distributed import get_ddp_info

_T_co = TypeVar('_T_co', covariant=True)
_TORCH_SAMPLERS = (RandomSampler, SubsetRandomSampler, WeightedRandomSampler)

# If we need resume-like reproducibility, then .step() if necessary
# (i.e. using BASE_SEED + epoch as seed for each epoch).
# Though this way, any alteration of epoch len will break reproducibility.
# This is useful when epoch len = dataset len, and sampling is done
# without replacement.

# If we only need from-scratch-like reproducibility, then we can drop .step()
# and rely on generator state.
# This way we can decouple on epoch len.
# Epoch transitions will be blurred, so the whole train loop will be like on
# the infinite dataset.
# Useful with replacement sampling and custom epoch len.


def generate_seed() -> int:
    return int(torch.empty((), dtype=torch.int64).random_().item())


class SamplerLike(Protocol, Generic[_T_co]):
    def __iter__(self) -> Iterator[_T_co]:
        ...

    def __len__(self) -> int:
        ...


class DdpSampler(Sampler[_T_co]):
    def __init__(self,
                 sampler: Sampler[_T_co] | SamplerLike[_T_co],
                 drop_last: bool = False) -> None:
        assert isinstance(sampler, Sized)
        self.base = sampler
        self.drop_last = drop_last
        self.seed = generate_seed()

        if (isinstance(sampler, _TORCH_SAMPLERS)
                and sampler.generator is None):
            # Enforce to use local generator
            sampler.generator = torch.Generator()

    def step(self, n: int = 1) -> None:
        self.seed += n

        # If torch random sampler, reseed its generator
        if isinstance(self.base, _TORCH_SAMPLERS):
            assert isinstance(self.base.generator, torch.Generator)
            self.base.generator.manual_seed(self.seed)

        # Non-sequential sampler, treat as random sampler
        # But as we don't know whether it uses generator reseed globally
        elif not isinstance(self.base, SequentialSampler):
            torch.manual_seed(self.seed)

    def __iter__(self) -> Iterator[_T_co]:
        self.step()

        if ddp := get_ddp_info():
            total = len(self) * ddp.world
            indices = iter(self.base)

            if (padding := total - len(self.base)) > 0:
                if len(self.base) < ddp.world:
                    indices = cycle(indices)
                else:
                    *head, = islice(indices, padding)
                    indices = chain(head, indices, head)

            return islice(indices, ddp.rank, total, ddp.world)

        return iter(self.base)

    def __len__(self) -> int:
        if ddp := get_ddp_info():
            if self.drop_last:
                return (len(self.base) + ddp.world - 1) // ddp.world
            return len(self.base) // ddp.world
        return len(self.base)



from __future__ import annotations

__all__ = [
    'Attention',
    'FeedForward',
    'MaxVitBlock',
    'MultiAxisAttention',
    'VitBlock',
]

import torch
import torch.nn.functional as F
from einops import rearrange
from einops.layers.torch import Rearrange
from packaging.version import Version
from torch import nn

from .aggregates import pre_norm
from .context import ConvCtx
from .convnets import mbconv
from .util import NameMixin, round8

_IS_TORCH_1_12 = Version('1.12') <= Version(torch.__version__) < Version('2.0')
_IS_TORCH_2X = Version(torch.__version__) >= Version('2.0')
_TORCH_MHA_AUTOCAST = True


class ReAttention(nn.Sequential):
    """Re-Attention from [DeepViT](https://arxiv.org/abs/2103.11886)"""
    def __init__(self, heads: int) -> None:
        super().__init__(
            Rearrange('b h i j -> b i j h'),
            nn.Linear(heads, heads, bias=False),
            nn.LayerNorm(heads),
            Rearrange('b i j h -> b h i j'),
        )
        nn.init.normal_(self[1].weight)


class Attention(NameMixin, nn.Module):
    """
    Multihead self-attention module (M-SA)
    from [ViT](https://openreview.net/pdf?id=YicbFdNTTy).

    Supports Re-attention mechanism
    from [DeepViT](https://arxiv.org/abs/2103.11886).
    """
    __constants__ = ['reattention']

    def __init__(self,
                 dim: int,
                 dim_head: int = 64,
                 dropout: float = 0.,
                 qkv_bias: bool = False,
                 reattention: bool = False):
        super().__init__()
        assert dim % dim_head == 0
        self.dim = dim
        self.heads = heads = dim // dim_head
        self.scale = dim_head ** -0.5
        self.dropout = dropout

        self.to_qkv = nn.Sequential(
            fc := nn.Linear(dim, 3 * dim, bias=qkv_bias),
            Rearrange('b n (split h d) -> split b h n d', h=heads, split=3),
        )
        nn.init.normal_(fc.weight, 0, (2 / (dim + dim_head)) ** .5)

        self.attend = nn.Sequential(
            nn.Softmax(-1),
            nn.Dropout(dropout, inplace=True),
        )
        self.reattention = reattention
        if reattention:
            self.attend.append(ReAttention(heads))

        self.to_out = nn.Sequential(
            Rearrange('b h n d -> b n (h d)'),
            fc := nn.Linear(dim, dim),
            nn.Dropout(dropout, inplace=True),
        )
        nn.init.xavier_normal_(fc.weight)

        self.name = f'{dim}, {heads=}'
        if qkv_bias:
            self.name += ', qkv_bias=True'
        if dropout:
            self.name += f', {dropout=}'
        if reattention:
            self.name += ', reattention=True'

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Optimized eval-only impl since Torch 1.12
        if _IS_TORCH_1_12 and not self.training and not self.reattention:
            in_w, in_b, out_w, out_b = (self.to_qkv[0].weight,
                                        self.to_qkv[0].bias,
                                        self.to_out[1].weight,
                                        self.to_out[1].bias)
            tensor_args = (x, in_w, in_b, out_w, out_b)
            if ((_TORCH_MHA_AUTOCAST or not torch.is_autocast_enabled())
                    and not (torch.is_grad_enabled() and any([  # noqa: C419
                        t.requires_grad for t in tensor_args
                    ]))):
                if torch.is_autocast_enabled():
                    # torch uses slowpath, but this allows it go fast
                    dtype = torch.get_autocast_gpu_dtype()
                    if in_b.dtype != dtype:
                        in_b = in_b.to(dtype)
                out, _ = torch._native_multi_head_attention(
                    x, x, x, self.dim, self.heads, in_w, in_b, out_w, out_b,
                    None, False)
                return out

        # b n dim -> b h n d
        qkv = self.to_qkv(x)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy

        # Use FLASH-attention (https://arxiv.org/abs/2205.14135)
        # and Memory-Efficient attention from XFormers
        # for PyTorch 2.x
        if _IS_TORCH_2X and not self.reattention:
            dropout = self.droupout if self.is_train else 0
            out = F.scaled_dot_product_attention(q, k, v, dropout_p=dropout)

        else:
            # Compute weights for each token
            dots = torch.einsum('bhid,bhjd -> bhij', q, k)
            attn = self.attend(dots * self.scale)

            # Remix tokens using weights
            out = torch.einsum('bhij,bhjd -> bhid', attn, v)

        # Restore shape, b h n d -> b n (h d)
        return self.to_out(out)


class _RelativePositionalBias(nn.Module):
    def __init__(self, heads: int, window_size: int) -> None:
        super().__init__()

        wdiff = 2 * window_size - 1
        self.bias = nn.Sequential(
            nn.Embedding(wdiff ** 2, heads),
            Rearrange('i j h -> h i j'),
        )

        axis = torch.arange(window_size)
        grid = torch.stack(torch.meshgrid(axis, axis, indexing='ij'), -1)
        pos = (
            rearrange(grid, 'i j c -> (i j) 1 c')
            - rearrange(grid, 'i j c -> 1 (i j) c'))
        pos -= pos.min()
        indices = pos @ torch.tensor([wdiff, 1])
        self.register_buffer('indices', indices, persistent=False)

    def forward(self) -> torch.Tensor:
        return self.bias(self.indices)


class MultiAxisAttention(NameMixin, nn.Module):
    """
    Multi-axis self-attention (Max-SA)
    from [MaxViT](https://arxiv.org/abs/2204.01697)
    """
    def __init__(self,
                 dim: int,
                 dim_head: int = 32,
                 dropout: float = 0.,
                 qkv_bias: bool = False,
                 window_size: int = 7):
        super().__init__()
        assert dim % dim_head == 0
        heads = dim // dim_head

        self.scale = dim_head ** -0.5
        self.to_qkv = nn.Sequential(
            nn.Linear(dim, dim * 3, bias=qkv_bias),
            Rearrange('... i (s h d) -> s ... h i d', s=3, h=heads),
        )
        self.bias = _RelativePositionalBias(heads, window_size)

        self.attend = nn.Sequential(
            nn.Softmax(-1),
            nn.Dropout(dropout, inplace=True),
        )
        self.to_out = nn.Sequential(
            Rearrange('... h i d -> ... i (h d)'),
            nn.Linear(dim, dim, bias=False),
            nn.Dropout(dropout, inplace=True),
        )
        self.name = f'{dim}, {heads=}, {window_size=}'
        if qkv_bias:
            self.name += ', qkv_bias=True'
        if dropout:
            self.name += f', {dropout=}'

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # ... i d -> ... i d, self-attention over i
        q, k, v = self.to_qkv(x).unbind(0)  # ... h i d

        sim = torch.einsum('... i d, ... j d -> ... i j', q, k) * self.scale
        sim += self.bias()
        attn = self.attend(sim)

        # aggregate
        out = torch.einsum('... i j, ... j d -> ... i d', attn, v)

        # combine heads out
        return self.to_out(out)


class FeedForward(NameMixin, nn.Sequential):
    def __init__(self, dim: int, ratio: float, dropout: float = 0.):
        dim_inner = round8(dim * ratio)
        super().__init__(
            nn.Linear(dim, dim_inner),
            nn.GELU(),
            nn.Dropout(dropout, inplace=True),
            nn.Linear(dim_inner, dim),
            nn.Dropout(dropout, inplace=True),
        )
        self.name = f'{dim}, {dim_inner=}'
        if dropout:
            self.name += f', {dropout=}'


# ----------------------------- complete blocks ------------------------------


class VitBlock(nn.Sequential):
    def __init__(self,
                 dim: int,
                 dim_head: int,
                 mlp_ratio: float = 4.,
                 dropout: float = 0.,
                 qkv_bias: bool = False,
                 reattn: bool = False):
        super().__init__(
            pre_norm(Attention(dim, dim_head, dropout, qkv_bias, reattn)),
            pre_norm(FeedForward(dim, mlp_ratio, dropout)),
        )


class MaxVitBlock(nn.Sequential):
    def __init__(self,
                 dim: int,
                 dim_head: int,
                 window: int,
                 stride: int = 1,
                 bn_ratio: float = 4.,
                 se_ratio: float = 0.25,
                 mlp_ratio: float = 4,
                 dropout: float = 0.,
                 qkv_bias: bool = False,
                 ctx: ConvCtx | None = None):
        super().__init__(
            # mbconv
            mbconv(dim, stride, bn_ratio, se_ratio, dropout, ctx),

            # block attention
            Rearrange('b d (x u) (y v) -> b x y (u v) d', u=window, v=window),
            pre_norm(
                MultiAxisAttention(dim, dim_head, dropout, qkv_bias, window)),
            pre_norm(FeedForward(dim, mlp_ratio, dropout)),
            Rearrange('b x y (u v) d -> b d (x u) (y v)', u=window, v=window),

            # grid attention
            Rearrange('b d (u x) (v y) -> b x y (u v) d', u=window, v=window),
            pre_norm(
                MultiAxisAttention(dim, dim_head, dropout, qkv_bias, window)),
            pre_norm(FeedForward(dim, mlp_ratio, dropout)),
            Rearrange('b x y (u v) d -> b d (u x) (v y)', u=window, v=window),
        )



from __future__ import annotations

__all__ = ['Sound']

from contextlib import ExitStack
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from queue import Queue
from threading import Event
from typing import Generic, TypeVar

import numpy as np
import numpy.typing as npt
from tqdm.auto import tqdm

from .. import chunked

_Scalar = TypeVar('_Scalar', bound=np.generic, covariant=True)


def _play(arr: np.ndarray,
          rate: int,
          blocksize: int = 1024,
          bufsize: int = 20) -> None:
    """Plays audio from array. Supports interruption via Crtl-C."""
    import sounddevice as sd

    q: Queue[np.ndarray | None] = Queue(bufsize)
    ev = Event()

    def callback(out: np.ndarray, *_) -> None:
        if (data := q.get()) is None:
            raise sd.CallbackAbort

        size = len(data)
        out[:size] = data
        if size < len(out):
            out[size:] = 0
            raise sd.CallbackStop

    stream = sd.OutputStream(
        rate, blocksize, callback=callback, finished_callback=ev.set)

    fmt = '{percentage:3.0f}% |{bar}| [{elapsed}<{remaining}]'
    blocks = chunked(arr, blocksize)

    with ExitStack() as s:
        s.enter_context(stream)  # Close stream
        s.callback(ev.wait)  # Wait for completion
        s.callback(q.put, None)  # Close queue

        for data in s.enter_context(
                tqdm(blocks, leave=False, smoothing=0, bar_format=fmt)):
            q.put(data)


@dataclass(repr=False, frozen=True)
class Sound(Generic[_Scalar]):
    """Wraps numpy.array to be playable as sound

    Parameters:
    - rate - sample rate to use for playback

    Usage:
    ```
    import numpy as np

    sound = Sound.load('test.flac')

    # Get properties
    rate: int = sound.rate
    dtype: np.dtype = sound.dtype

    # Could be played like:
    import sounddevice as sd
    sd.play(sound, sound.rate)

    # Or like this, if you need Ctrl-C support
    sound.play()

    # Extract underlying array
    raw = sound.raw

    # Same result
    raw = np.array(sound)
    ```
    """
    data: npt.NDArray[_Scalar]
    rate: int = 44_100

    def __post_init__(self):
        if self.data.ndim not in (1, 2):
            raise ValueError('Sound must be 1d (mono) or 2d array, '
                             f'got {self.data.shape}')
        if self.data.ndim == 1:
            object.__setattr__(self, 'data', self.data[:, None])
        if self.data.shape[-1] not in (1, 2):
            raise ValueError('Only mono/stereo is supported, '
                             f'got {self.channels} channels')
        if self.data.dtype not in ('i1', 'i2', 'i4', 'f4'):
            raise ValueError('Only int8/int16/int32/float32 sound dtype '
                             f'is supported. Got {self.data.dtype}')

    @property
    def channels(self) -> int:
        return self.data.shape[1]

    @property
    def duration(self) -> timedelta:
        return timedelta(seconds=self.data.shape[0] / self.rate)

    def __repr__(self) -> str:
        duration = self.duration
        channels = self.channels
        dtype = self.data.dtype
        return f'{type(self).__name__}({duration=!s}, {channels=}, {dtype=!s})'

    def __array__(self) -> npt.NDArray[_Scalar]:
        return self.data

    def play(self, blocksize=1024) -> None:
        """Plays audio from array. Supports interruption via Crtl-C."""
        _play(self.data, self.rate, blocksize=blocksize)

    @classmethod
    def load(cls, path: Path | str) -> Sound:
        _check_fmt(path)
        import soundfile

        data, rate = soundfile.read(path)
        return cls(data.astype('f4'), rate)

    def save(self, path: Path | str) -> None:
        _check_fmt(path)
        import soundfile

        soundfile.write(path, self.data, self.rate)


def _check_fmt(path: Path | str):
    fmt = Path(path).suffix.lower()
    if fmt not in ('.flac', '.wav'):
        raise ValueError(f'Only FLAC/WAV is supported, got {fmt}')



"""Make builtin iterators compatible with `len`"""
__all__ = ['apply']

import builtins
import functools
import math
import operator
from collections.abc import Iterable, Iterator, Sized
from dataclasses import dataclass
from itertools import (accumulate, count, cycle, islice, product, repeat,
                       starmap, tee, zip_longest)
from typing import Generic, Protocol, TypeVar, runtime_checkable

# -------------------------- iterable-size proxies ---------------------------

_T_co = TypeVar('_T_co', covariant=True)
_INF = float('inf')


@runtime_checkable
class SizedIterable(Sized, Iterable[_T_co], Protocol[_T_co]):
    ...


@dataclass(repr=False, frozen=True)
class _SizedIterable(Generic[_T_co]):
    _it: Iterable[_T_co]
    _size: int

    def __iter__(self) -> Iterator[_T_co]:
        return iter(self._it)

    def __len__(self) -> int:
        return self._size

    def __repr__(self) -> str:
        line = object.__repr__(self._it)
        line = line.removeprefix('<').removesuffix('>')
        return f'<{line} with {self._size} items>'


# --------------------------------- builtins ---------------------------------

len_hint = functools.singledispatch(builtins.len)

_tee: type = tee(())[0].__class__
_iterables: list[Iterable] = [
    '', b'',
    range(0), (), [], {}, {}.keys(), {}.values(), {}.items(),
    reversed(()),
    reversed([]),
    set(),
    frozenset()
]
_transparent_types: tuple[type, ...] = tuple(
    it.__iter__().__class__ for it in _iterables)
for _tp in _transparent_types:
    len_hint.register(_tp, operator.length_hint)


def _are_definitely_independent(iters):
    return (len({id(it) for it in iters}) == len(iters)
            and all(isinstance(it, _transparent_types) for it in iters))


@len_hint.register(zip)
def _len_zip(x):  # type: ignore[misc]
    _, iters = x.__reduce__()
    if not iters:
        return 0
    if len(iters) == 1:
        return len(iters[0])

    # Do not compute zip size when it's constructed from multiple iterables.
    # as there's currently no reliable way to check whether underlying
    # iterables are independent or not
    if _are_definitely_independent(iters):
        return min(map(len, iters))

    raise TypeError


@len_hint.register(map)
def _len_map(x):  # type: ignore[misc]
    _, (__fn, *iters) = x.__reduce__()
    if len(iters) == 1:
        return len(iters[0])

    # Same as for zip above
    if _are_definitely_independent(iters):
        return min(map(len, iters))

    raise TypeError


# --------------------------- itertools.infinite ---------------------------


@len_hint.register(count)
def _len_count(_):  # type: ignore[misc]
    return _INF


@len_hint.register(cycle)
def _len_cycle(x):  # type: ignore[misc]
    _, [iterable], (buf, pos) = x.__reduce__()
    if buf or len(iterable):
        return _INF
    return 0


@len_hint.register(repeat)
def _len_repeat(x):  # type: ignore[misc]
    _, (obj, *left) = x.__reduce__()
    return left[0] if left else _INF


# ---------------------------- itertools.finite ----------------------------


@len_hint.register(accumulate)
def _len_accumulate(x):  # type: ignore[misc]
    _, (seq, _fn), _total = x.__reduce__()
    return len(seq)


# @len_hint.register(chain)


@len_hint.register(islice)
def _len_islice(x):  # type: ignore[misc]
    _, (it, start, *stop_step), done = x.__reduce__()
    if not stop_step:
        return 0
    stop, step = stop_step
    total = len(it) + done
    stop = total if stop is None else min(total, stop)
    if math.isinf(stop):
        # range can't handle inf
        return _INF
    return len(range(start, stop, step))


@len_hint.register(starmap)
def _len_starmap(x):  # type: ignore[misc]
    _, (_fn, seq) = x.__reduce__()
    return len(seq)


@len_hint.register(_tee)
def _len_tee(x):
    _, [empty_tuple], (dataobject, pos) = x.__reduce__()
    _, (src, buf, none) = dataobject.__reduce__()
    return len(src) + len(buf) - pos


@len_hint.register(zip_longest)
def _len_zip_longest(x):  # type: ignore[misc]
    _, iters, _pad = x.__reduce__()
    if _are_definitely_independent(iters):
        return max(map(len, iters))
    raise TypeError


# -------------------------- itertools.combinatoric -------------------------


@len_hint.register(product)
def _len_product(x):  # type: ignore[misc]
    _, seqs, *pos = x.__reduce__()

    # Greedy caches all input iterables, no need to check interference
    lens = *map(len, seqs),
    total = functools.reduce(operator.mul, lens, 1)
    if not pos:
        return total

    strides = accumulate((1, ) + lens[:0:-1], operator.mul)
    offset = sum(map(operator.mul, strides, reversed(pos[0])))
    return total - offset - 1


# @len_hint.register(permutations)
# @len_hint.register(combinations)
# @len_hint.register(combinations_with_replacement)


def apply():
    builtins.len = len_hint



from __future__ import annotations

__all__ = ['Trainer']

from collections.abc import Callable, Iterable, Iterator
from contextlib import nullcontext
from dataclasses import InitVar, dataclass
from typing import Any

import torch
import torch.optim
from torch import nn
from tqdm.auto import tqdm

from .. import ichunked
from .. import metrics as m
from ._loader import _Loader
from .amp import Grads, get_grads
from .util import eval_


@dataclass
class Stage:
    net: nn.Module
    device: torch.device
    dtype: InitVar[torch.dtype | None]

    def __post_init__(self, dtype: torch.dtype | None):
        self._autocast: Any = (
            torch.autocast(self.device.type, dtype)
            if dtype in (torch.float16, torch.bfloat16) else nullcontext())

    def _move(self, x: torch.Tensor) -> torch.Tensor:
        return x.to(self.device, non_blocking=True)

    def __call__(self, loader: _Loader) -> Iterator[tuple[torch.Tensor, ...]]:
        raise NotImplementedError


@dataclass
class EvalStage(Stage):
    def _step(self, data: torch.Tensor,
              target: torch.Tensor) -> tuple[torch.Tensor, ...]:
        with self._autocast:
            out = self.net(self._move(data))

        return out, target

    def __call__(self, loader: _Loader) -> Iterator[tuple[torch.Tensor, ...]]:
        with eval_(self.net), torch.inference_mode():
            for data, target in loader:
                yield self._step(self._move(data), self._move(target))


@dataclass
class TrainStage(Stage):
    criterion: Callable[..., torch.Tensor]
    grads: Grads
    grad_steps: int = 1

    def _step(self, data: torch.Tensor,
              target: torch.Tensor) -> tuple[torch.Tensor, ...]:
        with self._autocast:
            out = self.net(self._move(data))
            loss = self.criterion(out, target)

        self.grads.backward(loss)
        return out.detach(), target

    def __call__(self, loader: _Loader) -> Iterator[tuple[torch.Tensor, ...]]:
        for batches in ichunked(loader, self.grad_steps):
            with self.grads:
                for data, target in batches:
                    yield self._step(self._move(data), self._move(target))
                # Clip norm here if needed


class Trainer:
    def __init__(self,
                 net: nn.Module,
                 opt: torch.optim.Optimizer,
                 criterion: Callable[..., torch.Tensor],
                 metrics: Iterable[m.Metric],
                 device: torch.device,
                 sched: torch.optim.lr_scheduler._LRScheduler | None = None,
                 dtype: torch.dtype | None = None,
                 grad_steps: int = 1) -> None:
        self.metrics = [m.Lambda(criterion, name='loss'), *metrics]

        grads = get_grads(opt, sched, dtype, max_retries=0)
        self.stages = (
            TrainStage(net, device, dtype, criterion, grads, grad_steps),
            EvalStage(net, device, dtype),
        )

    def _run(self, stage: Stage, loader: _Loader, pbar: tqdm) -> m.Scores:
        meter = m.compose(*self.metrics)
        scores = m.Scores()

        for out in stage(loader):
            scores = meter.send(out)
            pbar.set_postfix(scores.scalars)
            pbar.update()

        return scores

    def train(self, loader: _Loader, pbar: tqdm) -> m.Scores:
        return self._run(self.stages[0], loader, pbar)

    def eval(self, loader: _Loader, pbar: tqdm) -> m.Scores:
        return self._run(self.stages[1], loader, pbar)

    def run(self,
            train_loader: _Loader,
            eval_loader: _Loader,
            epochs: int = 1):
        for i in tqdm(range(1, 1 + epochs), smoothing=0):
            with tqdm(train_loader, desc='train', leave=False) as bar:
                tscalars = self.train(bar, bar).scalars

            with tqdm(eval_loader, desc='val', leave=False) as bar:
                vscalars = self.eval(bar, bar).scalars

            assert tscalars.keys() == vscalars.keys()
            tags = sorted(tscalars.keys() | vscalars.keys())

            # TODO: those lines should be moved outsize into loggers
            line = ','.join(
                f'{tag}: ' + '/'.join(f'{s[tag]:.3f}'
                                      for s in (tscalars, vscalars))
                for tag in tags)
            print(f'[{i:03d}] {line}')


# TODO: define Logger/Handler classes
# TODO:  i.e. Logger which handles tensors and scalars
# TODO:  and Handler flavours like StdoutHandler, TensorBoardHandler, etc.



"""
? At each moment each thread serve one executor.
+ Threads are reused over all executors.
+ Idle threads are stopped on _TIMEOUT.
? Most recently used threads used first (a.k.a. stack, more dropout).
"""
from __future__ import annotations

__all__ = ['ThreadQuota']

from collections import deque
from collections.abc import Callable
from concurrent.futures import Executor, Future
from concurrent.futures._base import LOGGER
from concurrent.futures.thread import _WorkItem
from queue import Empty, SimpleQueue
from threading import _register_atexit  # type: ignore[attr-defined]
from threading import Lock, Thread
from typing import TypeVar
from weakref import WeakKeyDictionary, WeakSet

# TODO: investigate hangups when _TIMEOUT <= .01
_TIMEOUT = 1
_MIN_IDLE = 10

# ------------------------------- generics -----------------------------------

_T = TypeVar('_T')


def _safe_call(fn: Callable[..., _T], *args, **kwargs) -> _T | None:
    try:
        return fn(*args, **kwargs)
    except (Empty, IndexError, ValueError):
        return None


# ------------------------------ implementation ------------------------------

_Pipe = SimpleQueue['ThreadQuota | None']

_shutdown = False
_shutdown_lock = Lock()
_executors = WeakSet['ThreadQuota']()
_workers = WeakKeyDictionary[Thread, _Pipe]()
_idle = deque[_Pipe]()


def _python_exit():
    global _shutdown  # noqa: PLW0603
    with _shutdown_lock:
        _shutdown = True

    for e in _executors:
        e.shutdown(cancel_futures=True)
    *items, = _workers.items()
    for _, q in items:
        q.put(None)
    for w, _ in items:
        w.join()


_register_atexit(_python_exit)


def _worker(q: _Pipe) -> None:
    try:
        while executor := _safe_call(q.get, timeout=_TIMEOUT):
            while work_item := _safe_call(executor._work_queue.popleft):
                work_item.run()  # Process task
                if _shutdown:
                    return

            executor._idle.append(1)  # Decrease worker usage for executor
            _idle.append(q)  # Mark worker as idle, LIFO/stack
            if _shutdown:
                return

    except BaseException:  # noqa: BLE001
        LOGGER.critical('Exception in worker', exc_info=True)
    finally:
        if _TIMEOUT:
            _safe_call(_idle.remove, q)  # Omit when '_idle' tracks weakrefs


class ThreadQuota(Executor):
    __slots__ = ('_work_queue', '_idle', '_shutdown_lock', '_shutdown')

    def __init__(self, max_workers: int):
        self._work_queue = deque[_WorkItem]()
        self._idle = [1] * max_workers  # semaphore

        self._shutdown_lock = Lock()
        self._shutdown = False

        with _shutdown_lock:
            _executors.add(self)

    def submit(self, fn, /, *args, **kwargs):
        with self._shutdown_lock, _shutdown_lock:
            if self._shutdown or _shutdown:
                raise RuntimeError('cannot schedule futures after shutdown')

            f = Future()  # type: ignore[var-annotated]
            self._work_queue.append(_WorkItem(f, fn, args, kwargs))

            if _safe_call(self._idle.pop):  # Pool is not maximized yet
                if not (q := _safe_call(_idle.pop)):
                    q = _Pipe()
                    w = Thread(target=_worker, args=(q, ))
                    w.start()
                    _workers[w] = q
                q.put(self)

            return f

    def shutdown(self, wait=True, *, cancel_futures=False):
        with self._shutdown_lock:
            if self._shutdown:
                return
            self._shutdown = True

            if cancel_futures:
                while work_item := _safe_call(self._work_queue.pop):
                    work_item.future.cancel()

            if not _TIMEOUT:
                # Keep at most 25% of workers idle
                while (len(_idle) > max(len(_workers) / 4, _MIN_IDLE)
                       and (q := _safe_call(_idle.popleft))):
                    q.put(None)



from .classes import (BitFlipNoise, ChannelMix, ChannelShuffle, CutOut,
                      DegradeJpeg, DegradeQuality, Elastic, FlipAxis, HsvShift,
                      LumaJitter, MaskDropout, MultiNoise, WarpAffine)
from .core import Chain, Transform
from .functional import grid_shuffle

__all__ = [
    'BitFlipNoise', 'Chain', 'ChannelMix', 'ChannelShuffle', 'CutOut',
    'DegradeJpeg', 'DegradeQuality', 'Elastic', 'FlipAxis', 'HsvShift',
    'LumaJitter', 'MaskDropout', 'MultiNoise', 'Transform', 'WarpAffine',
    'grid_shuffle'
]



from itertools import islice

from tqdm.auto import tqdm

import glow


def iter_fibs():
    prev, cur = 0, 1
    while True:
        prev, cur = cur, prev + cur
        yield cur


@glow.time_this
def fibs(n):
    gen = islice(iter_fibs(), n)
    gen = tqdm(gen)
    return max(x.bit_length() for x in gen)


def test_fibs():
    assert fibs(100_000) == 69_424



import random
from unittest.mock import MagicMock

import pytest

import glow


def test_as_is():
    load = MagicMock(side_effect=lambda x: x)

    @glow.memoize(8192, batched=True)
    def load_batch(xs):
        return [load(x) for x in xs]

    for _ in range(5):
        numbers = random.choices(range(20), k=5)
        results = load_batch(numbers)
        assert numbers == results

    called_with = (x for (x, ), _ in load.call_args_list)
    assert sorted(called_with) == sorted(load_batch.cache)


@pytest.mark.parametrize('max_workers', [2, 4] * 10)
def test_thread_safe(max_workers):
    @glow.memoize(8192, batched=True)
    def load_batch(xs):
        return [*xs]

    numbers = random.choices(range(100), k=100)
    nchunks = glow.chunked(numbers, 5)
    rchunks = glow.map_n(load_batch, nchunks, max_workers=max_workers)
    assert [r for c in rchunks for r in c] == numbers



__all__ = ['accuracy_', 'auroc', 'average_precision', 'dice']

import torch

from .base import to_index, to_prob


def accuracy_(pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
    # TODO: Add docs
    _, pred, true = to_index(pred, true)
    return (true == pred).double().mean()


def dice(pred: torch.Tensor,
         true: torch.Tensor,
         macro: bool = True) -> torch.Tensor:
    # TODO: Add docs
    c, pred, true = to_index(pred, true)

    def _dice(pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
        true = true.view(-1)
        pred = pred.view(-1)
        tp, t, p = (
            x.bincount(minlength=c).clamp_(1).double()
            for x in (true[true == pred], true, pred))
        return 2 * tp / (t + p)

    if macro:
        return _dice(pred, true)

    b = pred.shape[0]
    *scores, = map(_dice, pred.view(b, -1).unbind(), true.view(b, -1).unbind())
    return torch.mean(torch.stack(scores), dim=0)


def _rankdata(ten: torch.Tensor) -> torch.Tensor:
    sorter = ten.argsort()
    ten = ten[sorter]

    diff = torch.cat([torch.tensor([True]), ten[1:] != ten[:-1]])
    # diff = np.r_[True, ten[1:] != ten[:-1]]

    dense = diff.cumsum(0)[sorter.argsort()]

    diff = diff.nonzero(as_tuple=False).view(-1)
    count = torch.cat([diff, torch.tensor([diff.numel()])])
    # count = np.r_[diff.nonzero(diff).view(-1), diff.numel()]

    return 0.5 * (count[dense] + count[dense - 1] + 1)


def _binary_metric(fn):
    """Applies specified function only on probabilities of indexed class"""
    def call(pred: torch.Tensor,
             true: torch.Tensor,
             index: int = 0) -> torch.Tensor:
        c, probs, targets = to_prob(pred, true)
        assert 0 <= index < c
        return fn(probs[:, index].view(-1), (targets == index).view(-1))

    return call


@_binary_metric
def auroc(pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
    n = true.numel()
    n_pos = true.sum()

    r = _rankdata(pred)
    total = n_pos * (n - n_pos)
    return (r[true == 1].sum() - n_pos * (n_pos + 1) // 2) / float(total)


@_binary_metric
def average_precision(pred: torch.Tensor, true: torch.Tensor) -> torch.Tensor:
    n = true.numel()
    n_pos = true.sum()

    true = true[torch.argsort(pred)].flipud()
    weights = torch.arange(1, n + 1).float().reciprocal()
    precision = true.cumsum(0).float()
    return torch.einsum('i,i,i', true.float(), precision, weights) / n_pos



__all__ = ['AdamW', 'RAdam', 'SGDW']

import torch
from torch.optim import optimizer


class _OptimizerBase(optimizer.Optimizer):
    _step = 0

    def __getstate__(self):
        return {**super().__getstate__(), '_step': self._step}

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        self._step += 1
        for group in self.param_groups:
            args = self._update_group(group)
            for p in group['params']:
                if p.grad is None:
                    continue
                if p.grad.is_sparse:
                    raise RuntimeError('Sparse grads are not supported')
                with torch.no_grad():
                    self._do_step(p, group, self.state[p], *args)

        return loss

    def _update_group(self, group) -> tuple:
        return ()

    def _do_step(self, p, group, state, *args):
        raise NotImplementedError


class SGDW(_OptimizerBase):
    """Implements stochastic gradient descent (optionally with momentum).

    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf
    """
    def __init__(self,
                 params,
                 lr=0.003,
                 momentum=0,
                 dampening=0,
                 weight_decay=0,
                 nesterov=False):
        assert lr >= 0.0
        assert momentum >= 0.0
        assert weight_decay >= 0.0
        defaults = {
            'lr': lr,
            'momentum': momentum,
            'dampening': dampening,
            'weight_decay': weight_decay,
            'nesterov': nesterov,
        }
        assert (not nesterov or (momentum > 0 and dampening == 0)
                ), 'Nesterov momentum requires a momentum and zero dampening'
        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def _do_step(self, p, group, state):
        if group['weight_decay'] != 0:
            p.mul_(1 - group['lr'] * group['weight_decay'])

        momentum = group['momentum']
        grad = p.grad
        if momentum != 0:
            if state:
                grad = state['exp_avg']
                grad.mul_(momentum).add_(p.grad, alpha=1 - group['dampening'])
            else:
                grad = state['exp_avg'] = p.grad.clone().detach_()

            if group['nesterov']:
                grad = p.grad.add(grad, alpha=momentum)

        p.add_(grad, alpha=-group['lr'])


class AdamW(_OptimizerBase):
    r"""Implements AdamW algorithm.

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _Decoupled Weight Decay Regularization:
        https://arxiv.org/abs/1711.05101
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """
    def __init__(self,
                 params,
                 lr=1e-3,
                 betas=(0.9, 0.999),
                 eps=1e-8,
                 weight_decay=1e-2,
                 amsgrad=False) -> None:
        assert lr >= 0.0
        assert eps >= 0.0
        for i, beta in enumerate(betas):
            assert 0.0 <= beta < 1.0, f'Invalid beta at index {i}: {betas}'
        defaults = {
            'lr': lr,
            'betas': betas,
            'eps': eps,
            'weight_decay': weight_decay,
            'amsgrad': amsgrad,
        }
        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    def _update_group(self, group) -> tuple:
        beta1, beta2 = group['betas']
        bias_correction1 = 1 - beta1 ** self._step
        bias_correction2 = 1 - beta2 ** self._step
        return group['lr'] * (bias_correction2 ** 0.5) / bias_correction1,

    def _do_step(self, p, group, state, step_size):
        amsgrad = group['amsgrad']
        if not state:
            state['exp_avg'] = torch.zeros_like(p)
            state['exp_avg_sq'] = torch.zeros_like(p)
            if amsgrad:
                state['max_exp_avg_sq'] = torch.zeros_like(p)

        exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
        beta1, beta2 = group['betas']

        exp_avg.mul_(beta1).add_(p.grad, alpha=1 - beta1)
        exp_avg_sq.mul_(beta2).addcmul_(p.grad, p.grad, value=1 - beta2)

        if group['weight_decay'] != 0:
            p.mul_(1 - group['lr'] * group['weight_decay'])

        if amsgrad:
            max_exp_avg_sq = state['max_exp_avg_sq']
            torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
            denom = max_exp_avg_sq.sqrt().add_(group['eps'])
        else:
            denom = exp_avg_sq.sqrt().add_(group['eps'])

        p.addcdiv_(-step_size, exp_avg, denom)


class RAdam(_OptimizerBase):
    def __init__(self,
                 params,
                 lr=1e-3,
                 betas=(0.9, 0.999),
                 weight_decay=0.0,
                 decay_to_sgd=True) -> None:
        self._decay_to_sgd = decay_to_sgd
        defaults = {'lr': lr, 'betas': betas, 'weight_decay': weight_decay}
        super().__init__(params, defaults)

    def _update_group(self, group):
        beta1, beta2 = group['betas']
        bias_correction1 = 1 - beta1 ** self._step
        bias_correction2 = 1 - beta2 ** self._step

        beta2_t = beta2 ** self._step
        n_sma_max = 2 / (1 - beta2) - 1
        n_sma = n_sma_max - 2 * self._step * beta2_t / bias_correction2

        # more conservative since it's an approximated value
        # variance is not tractable
        if n_sma < 5:
            return False, (1 / bias_correction1)

        k = (n_sma - 4) * (n_sma - 2) / n_sma
        k_max = (n_sma_max - 4) * (n_sma_max - 2) / n_sma_max
        step_size = ((1 - beta2_t) * k / k_max) ** 0.5 / bias_correction1
        return True, step_size

    def _do_step(self, p, group, state, is_tractable, step_size):
        if not state:
            state['exp_avg'] = torch.zeros_like(p)
            state['exp_avg_sq'] = torch.zeros_like(p)

        exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
        beta1, beta2 = group['betas']

        exp_avg.mul_(beta1).add_(p.grad, alpha=1 - beta1)
        exp_avg_sq.mul_(beta2).addcmul_(p.grad, p.grad, value=1 - beta2)

        if group['weight_decay'] != 0:
            p.data.mul_(1 - group['weight_decay'] * group['lr'])

        if is_tractable:
            denom = exp_avg_sq.sqrt().add_(1e-8)
            p.addcdiv_(exp_avg, denom, value=-step_size * group['lr'])
        elif self._decay_to_sgd:
            p.add_(exp_avg, alpha=-step_size * group['lr'])



__all__ = ['make_key']

from collections.abc import Hashable

_KWD_MARK = object()


class _HashedSeq(list):  # List is mutable, that's why not NamedTuple
    __slots__ = 'hashvalue',

    def __init__(self, tup: tuple):
        self[:] = tup
        self.hashvalue = hash(tup)  # Memorize hash

    def __hash__(self):
        return self.hashvalue


def make_key(*args, **kwargs) -> Hashable:
    """Copied from functools._make_key, as private function"""
    if len(args) == 1 and isinstance(args[0], (int, str)):
        return args[0]
    if kwargs:
        args = sum(kwargs.items(), (*args, _KWD_MARK))
    return _HashedSeq(args)



from collections.abc import Callable
from contextlib import AbstractContextManager
from typing import TypeVar, overload

_F = TypeVar('_F', bound=Callable)


def memprof(name_or_callback: str | Callable[[float], object] | None = ...,
            /) -> AbstractContextManager[None]:
    ...


def timer(name_or_callback: str | Callable[[int], object] | None = ...,
          time: Callable[[], int] = ...,
          /) -> AbstractContextManager[None]:
    ...


@overload
def time_this(fn: _F, /, *, name: str | None = ...) -> _F:
    ...


@overload
def time_this(*, name: str | None = ...) -> Callable[[_F], _F]:
    ...



from __future__ import annotations

__all__ = ['register_post_import_hook', 'when_imported']

import sys
from collections.abc import Callable
from importlib import abc, util
from threading import RLock
from typing import Any, TypeVar

_Hook = Callable[[Any], object]
_HookVar = TypeVar('_HookVar', bound=_Hook)

_INITIALIZED = False
_LOCK = RLock()
_HOOKS: dict[str, list[_Hook]] = {}


class _ImportHookChainedLoader(abc.Loader):
    def __init__(self, loader):
        self.loader = loader

    def _set_loader(self, module):
        undefined = object()
        if getattr(module, '__loader__', undefined) in (None, self):
            try:
                module.__loader__ = self.loader
            except AttributeError:
                pass

        if ((spec := getattr(module, '__spec__', None)) is not None
                and getattr(spec, 'loader', None) is self):
            spec.loader = self.loader

    def create_module(self, spec):
        return self.loader.create_module(spec)

    def exec_module(self, module):
        self._set_loader(module)
        self.loader.exec_module(module)

        name: str | None = getattr(module, '__name__', None)
        with _LOCK:
            hooks = _HOOKS.pop(name, [])  # type: ignore[arg-type]
        for hook in hooks:
            hook(module)


class _ImportHookFinder(abc.MetaPathFinder, set[str]):
    def find_spec(self, fullname: str, path, target=None):
        with _LOCK:
            if fullname not in _HOOKS or fullname in self:
                return None

        self.add(fullname)
        try:
            if ((spec := util.find_spec(fullname)) and (loader := spec.loader)
                    and not isinstance(loader, _ImportHookChainedLoader)):
                spec.loader = _ImportHookChainedLoader(loader)
                return spec
        finally:
            self.remove(fullname)
        return None


def register_post_import_hook(hook: _Hook, name: str) -> None:
    """Register a new post import hook for the target module name.

    This will result in a proxy callback being registered which will defer
    loading of the specified module containing the callback function until
    required.

    Simplified version of wrapt.register_post_import_hook.
    """
    with _LOCK:
        global _INITIALIZED  # noqa: PLW0603
        if not _INITIALIZED:
            _INITIALIZED = True
            sys.meta_path.insert(0, _ImportHookFinder())

        if (module := sys.modules.get(name)) is not None:
            hook(module)
        else:
            _HOOKS.setdefault(name, []).append(hook)


def when_imported(name: str) -> Callable[[_HookVar], _HookVar]:
    """
    Decorator for marking that a function should be called as a post
    import hook when the target module is imported.

    Simplified version of wrapt.when_imported.
    """
    def wrapper(hook: _HookVar) -> _HookVar:
        register_post_import_hook(hook, name)
        return hook

    return wrapper



# flake8: noqa

import glow.nn as gnn

optD, optG = gnn.amp_init_opt([netD, netG], [optD, optG])


def step():
    netD.train()
    with optD:
        errD_real = criterion(netD(real), real_label)
        optD.backward(errD_real)

        fake = netG(noise)
        errD_fake = criterion(netD(fake.detach()), fake_label)
        optD.backward(errD_fake)

    netD.eval()
    with optG:
        errG = criterion(netD(fake), real_label)
        optG.backward(errG)



from itertools import count, islice

import torch
from torch.utils import data

from glow.nn import get_loader


class Dataset(data.Dataset):
    def __getitem__(self, index):
        return index

    def __len__(self):
        return 5


class PassDataset(data.Dataset):
    def __getitem__(self, index):
        return index

    def __len__(self):
        return 1


class PassTensorDataset(data.Dataset):
    def __getitem__(self, index):
        return torch.as_tensor([index])

    def __len__(self):
        return 1


class Sampler(data.Sampler):
    def __init__(self, n):
        self.n = n
        self.count = count()

    def __iter__(self):
        return iter(islice(self.count, self.n))

    def __len__(self):
        return self.n


def test_loader():
    # No batch, no collate, no tensors
    loader = get_loader(Dataset())
    assert len(loader) == 5
    assert [*loader] == [0, 1, 2, 3, 4]

    # Batch -> tensors
    loader2 = loader.batch(1)
    assert len(loader2) == 5
    assert [x.tolist() for x in loader2] == [[0], [1], [2], [3], [4]]

    loader2 = loader.batch(2)
    assert len(loader2) == 3
    assert [x.tolist() for x in loader2] == [[0, 1], [2, 3], [4]]


def test_pass_loader():
    loader = get_loader(PassDataset()).shuffle(Sampler(5)).batch(1)
    assert len(loader) == 5
    assert [x.tolist() for x in loader] == [[0], [1], [2], [3], [4]]

    # 2nd epoch
    assert len(loader) == 5
    assert [x.tolist() for x in loader] == [[5], [6], [7], [8], [9]]


def test_tensor_loader():
    loader = get_loader(PassTensorDataset()).shuffle(Sampler(5)).batch(1)
    assert len(loader) == 5
    assert [x.tolist() for x in loader] == [[[0]], [[1]], [[2]], [[3]], [[4]]]

    assert len(loader) == 5
    assert [x.tolist() for x in loader] == [[[5]], [[6]], [[7]], [[8]], [[9]]]



from .base import Lambda, Metric, Scores, Staged, compose, to_index, to_prob
from .confusion import (Confusion, ConfusionGrad, accuracy, accuracy_balanced,
                        iou, kappa, kappa_quadratic_weighted)
from .raw import accuracy_, auroc, average_precision, dice

__all__ = [
    'Confusion', 'ConfusionGrad', 'Lambda', 'Metric', 'Scores', 'Staged',
    'accuracy', 'accuracy_', 'accuracy_balanced', 'auroc', 'average_precision',
    'compose', 'dice', 'iou', 'kappa', 'kappa_quadratic_weighted', 'to_index',
    'to_prob'
]



from __future__ import annotations

__all__ = [
    'detach_', 'device', 'dump_to_onnx', 'eval_', 'frozen', 'inference',
    'materialize', 'param_count', 'profile'
]

import functools
from collections.abc import Callable, Iterable, Iterator, Mapping
from contextlib import contextmanager
from io import BytesIO
from itertools import islice
from typing import Any, TypeVar, cast

import numpy as np
import torch
from torch import nn

from .. import si
from .modules.lazy import _materialize_cls

_T = TypeVar('_T')
_F = TypeVar('_F', bound=Callable[..., Iterator])


def _apply(xs: _T, fn: Callable[[torch.Tensor], Any]) -> _T:
    if isinstance(xs, torch.Tensor):
        return fn(xs)

    if isinstance(xs, (str, bytes, np.ndarray)):
        return xs  # type: ignore

    if isinstance(xs, tuple) and hasattr(xs, '_fields'):  # namedtuple
        return type(xs)(*(_apply(x, fn) for x in xs))  # type: ignore
    if isinstance(xs, Mapping):
        return dict(_apply(kv, fn) for kv in xs.items())  # type: ignore
    if isinstance(xs, Iterable):
        return type(xs)(_apply(x, fn) for x in xs)  # type: ignore
    return xs


def device() -> torch.device:
    """Gets current device, including CPU"""
    return torch.device(f'cuda:{torch.cuda.current_device()}' if torch.cuda
                        .is_available() else 'cpu')


def param_count(module: nn.Module) -> int:
    """Count of parameters in net, both training and not"""
    params = {p for p in module.parameters() if not nn.parameter.is_lazy(p)}
    return si(sum(p.numel() for p in params))


@contextmanager
def eval_(module: nn.Module) -> Iterator[None]:
    """
    Switches all children to eval mode.
    Restores train/eval distribution at exit.
    """
    were_train = {m for m in module.modules() if m.training}
    try:
        module.eval()
        yield
    finally:
        for m in module.modules():
            if m in were_train:
                m.training = True  # Don't call .train() as it's recursive.


@contextmanager
def detach_(module: nn.Module) -> Iterator[None]:
    """Prevents module from changing its parameters.

    Forbids autograd to record operations on parameters in this module, thus
    excluding them from gradient computation.

    This method is helpful for freezing part of the module for finetuning or
    training parts of a model individually (e.g., GAN training).

    NEITHER disable gradient flow NOR prevents buffers to change.
    """
    required_grad = {
        p.detach_() for p in module.parameters()
        if not nn.parameter.is_lazy(p) and p.requires_grad
    }
    try:
        yield
    finally:
        for p in required_grad:
            p.requires_grad_()


@contextmanager
def frozen(module: nn.Module) -> Iterator[None]:
    """Blocks module from changing state of its parameters and buffers.

    Switches all children to eval mode and detaches all parameters.
    DOES NOT disable gradient flow.
    """
    with eval_(module), detach_(module):
        yield


@contextmanager
def inference(module: nn.Module) -> Iterator[None]:
    """Enables inference mode for module.

    Switches all children to eval mode.
    Disables gradient flow.

    All the tensors created in this mode are marked as inference,
    and they are NOT COMPATIBLE WITH AUTOGRAD AT ALL
    (used in JIT, backward, etc.).

    DON'T use this mode to initialize lazy modules.
    """
    with eval_(module), torch.inference_mode():
        yield


# ----------------------------- profile CUDA ops -----------------------------


def profile(fn: _F) -> _F:
    """Decorator to profile CUDA ops. Use with `nvprof`

    Use in script launched via:
    ```bash
    nvprof --profile-from-start off -o trace.prof -- python main.py
    ```
    Usage:
    >>> @profile
    ... def train_loop():
    ...     for data in loader:
    ...         yield step(data)

    """
    def wrapper(*args, **kwargs):
        results = fn(*args, **kwargs)
        with torch.cuda.profiler.profile():
            yield from islice(results, 1)
            with torch.autograd.profiler.emit_nvtx():
                yield from results

    return cast(_F, functools.update_wrapper(wrapper, fn))


def dump_to_onnx(model: nn.Module,
                 *shapes: tuple[int, ...],
                 device: str | torch.device = 'cpu') -> bytes:
    """Converts model to ONNX graph, represented as bytes

    Parameters:
    - model - torch.nn.Module to convert
    - shapes - Shapes of input data, all except batch dimension

    Example usage:
    >>> module = torch.nn.Linear(4, 4)
    >>> bytes_ = dump_to_onnx(module, [4])

    To restore graph:
    >>> from onnxruntime import backend
    >>> rep = backend.prepare(bytes_or_filename, device='cpu')
    >>> rep.run([np.zeros(4, 4)])[0]

    """
    dynamic_axes = {
        f'inp_{i}': {
            0: 'batch',
            **{dim: f'inp_{i}_dim_{dim}' for dim in range(2, 1 + len(shape))}
        } for i, shape in enumerate(shapes)
    }
    buf = BytesIO()
    torch.onnx.export(
        model.to(device).eval(),
        tuple(
            torch.rand(1, *s, requires_grad=True, device=device)
            for s in shapes),
        buf,
        input_names=[*dynamic_axes],
        dynamic_axes=dynamic_axes,
        opset_version=13,
        do_constant_folding=True)
    return buf.getvalue()


# --------------------------- fix for lazy module ----------------------------


def materialize(model: nn.Module, *args, **kwargs):
    """
    Materialize all the lazy modules within model.
    Safely call forward() if args or kwargs are passed.
    """
    lazy = {
        name: m for name, m in model.named_modules()
        if isinstance(m, nn.modules.lazy.LazyModuleMixin)
    }
    if not lazy:
        return

    uninitialized = {
        name: m for name, m in lazy.items()
        if m.has_uninitialized_params()  # type: ignore
    }
    if not uninitialized:  # Complete initialization without forward() call
        for m in lazy.values():
            _materialize_cls(m)  # type: ignore
        return

    if args or kwargs:  # Initialize from forward() call
        with eval_(model), torch.no_grad():
            model(*args, **kwargs)
        return

    raise RuntimeError(
        'Found uninitialized lazy modules but no example inputs were passed '
        'to initialize them:\n'
        f'{[*uninitialized]}')



from __future__ import annotations

__all__ = [
    'Lambda', 'Metric', 'Scores', 'Staged', 'compose', 'to_index',
    'to_index_sparse', 'to_prob', 'to_prob_sparse'
]

from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Sequence
from dataclasses import dataclass, field
from itertools import count
from typing import Protocol, overload

import torch

from .. import coroutine


@dataclass(frozen=True)
class Scores:
    scalars: dict[str, float | int] = field(default_factory=dict)
    tensors: dict[str, torch.Tensor] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, mapping: dict[str, torch.Tensor]) -> Scores:
        obj = cls()
        for k, v in mapping.items():
            if v.numel() == 1:
                obj.scalars[k] = v.item()
            else:
                obj.tensors[k] = v
        return obj


class MetricFn(Protocol):
    def __call__(self, pred, true) -> torch.Tensor:
        ...


class Metric(ABC):
    """Base class for metric"""
    @abstractmethod
    def __call__(self, pred, true) -> torch.Tensor:
        raise NotImplementedError

    @abstractmethod
    def collect(self, state: torch.Tensor) -> dict[str, torch.Tensor]:
        raise NotImplementedError


class Lambda(Metric):
    """Wraps arbitary loss function to metric"""
    fn: MetricFn

    @overload
    def __init__(self, fn: Callable, name: str):
        ...

    @overload
    def __init__(self, fn: MetricFn, name: None = ...):
        ...

    def __init__(self, fn, name=None):
        self.fn = fn
        self.name = fn.__name__ if name is None else name

    def __call__(self, pred, true) -> torch.Tensor:
        return self.fn(pred, true)

    def collect(self, state: torch.Tensor) -> dict[str, torch.Tensor]:
        return {self.name: state}


class Staged(Metric):
    """Makes metric a "producer": applies multiple functions to its "state" """
    def __init__(self, **funcs: Callable[[torch.Tensor], torch.Tensor]):
        self.funcs = funcs

    def collect(self, state: torch.Tensor) -> dict[str, torch.Tensor]:
        return {key: fn(state) for key, fn in self.funcs.items()}


def to_index(pred: torch.Tensor,
             true: torch.Tensor) -> tuple[int, torch.Tensor, torch.Tensor]:
    """
    Convert `pred` of logits with shape [B, C, ...] to [B, ...] of indices,
    i.e. tensors of long.
    """
    assert pred.shape[0] == true.shape[0]
    assert pred.shape[2:] == true.shape[1:]

    c = pred.shape[1]
    pred = pred.argmax(dim=1)

    return c, pred, true


def to_index_sparse(
        pred: torch.Tensor,
        true: torch.Tensor) -> tuple[int, torch.Tensor, torch.Tensor]:
    """
    Convert `pred` of logits with shape [B, C, ...] to [B, ...] of indices,
    i.e. tensors of long. Drops bad indices.
    Result is flattened.
    """
    c, pred, true = to_index(pred, true)

    pred = pred.view(-1)
    true = true.view(-1)

    mask = (true >= 0) & (true < c)
    return c, pred[mask], true[mask]


def to_prob(pred: torch.Tensor,
            true: torch.Tensor) -> tuple[int, torch.Tensor, torch.Tensor]:
    """
    Convert `pred` of logits with shape [B, C, ...] to probs,
    i.e. tensors of float32.
    """
    assert pred.shape[0] == true.shape[0]
    assert pred.shape[2:] == true.shape[1:]
    c = pred.shape[1]

    with torch.autocast('cuda'):  # Softmax is always fp32
        pred = pred.softmax(dim=1)

    assert pred.dtype == torch.float32
    return c, pred, true


def to_prob_sparse(
        pred: torch.Tensor,
        true: torch.Tensor) -> tuple[int, torch.Tensor, torch.Tensor]:
    """
    Convert `pred` of logits with shape [B, C, ...] to probs,
    i.e. tensors of float.
    Drops bad indices, i.e. those that are out of range(C).
    Results have shape of [N, C] and [N].
    """
    c, pred, true = to_prob(pred, true)

    b = true.shape[0]
    true = true.view(-1)  # (b n)
    pred = pred.view(b, c, -1).permute(0, 2, 1).view(-1, c)  # (b n) c

    mask = (true >= 0) & (true < c)
    return c, pred[mask], true[mask]


@coroutine
def _batch_averaged(
    fn: Metric
) -> Generator[dict[str, torch.Tensor], Sequence[torch.Tensor], None]:
    assert isinstance(fn, Metric)
    args = yield {}
    state = fn(*args)
    for n in count(2):
        args = yield fn.collect(state)
        state.lerp_(fn(*args), 1 / n)


@coroutine
def compose(*fns: Metric) -> Generator[Scores, Sequence[torch.Tensor], None]:
    updates = *(_batch_averaged(fn) for fn in fns),
    args = yield Scores()
    while True:
        scores: dict[str, torch.Tensor] = {}
        for u in updates:
            scores |= u.send(args)
        args = yield Scores.from_dict(scores)



from __future__ import annotations

__all__ = ['countable', 'mangle', 'repr_as_obj', 'si', 'si_bin']

from collections import Counter
from collections.abc import Callable
from typing import TypeVar, cast

from wrapt import ObjectProxy


def mangle() -> Callable[[str], str | None]:
    """Appends number to already seen strings, making them distinct

    >>> mangled = mangle()
    >>> mangled('a')
    'a'
    >>> mangled('b')
    'b'
    >>> mangled('a')
    'a:1'
    """
    store = Counter[str]()

    def call(name: str) -> str | None:
        if name is None:
            return None

        seen = store[name]
        store[name] += 1
        if not seen:
            return name
        return f'{name}:{seen}'

    return call


def countable() -> Callable[[object], int]:
    """Accumulates and enumerates objects. Readable alternative to id().

    >>> id_ = countable()
    >>> id_('a')
    0
    >>> id_('b')
    1
    >>> id_('a')
    0
    """
    instances: dict[int, int] = {}
    return lambda obj: instances.setdefault(id(obj), len(instances))


def repr_as_obj(d: dict) -> str:
    """Returns pretty representation of dict.

    >>> repr_as_obj({'a': 1, 'b': 2})
    'a=1, b=2'
    """
    return ', '.join(f'{key}={value!r}' for key, value in d.items())


# ----------------------- number type with pretty repr -----------------------

_IntOrFloat = TypeVar('_IntOrFloat', int, float)
_PREFIXES = 'qryzafpnum kMGTPEZYRQ'
_PREFIXES_BIN = _PREFIXES[_PREFIXES.index(' '):].upper()


def _find_prefix(value: float | int, base: int,
                 prefixes: str) -> tuple[float, str]:
    threshold = base - 0.5
    origin = prefixes.find(' ') + 1
    value *= base ** origin

    for prefix in prefixes:
        value /= base
        if -threshold < value < threshold:
            return value, prefix

    return value, prefixes[-1]


def _num_repr(value: float | int, si: bool = True) -> str:
    if value == 0:
        return '0' if si else '0B'

    base, prefixes = (1000, _PREFIXES) if si else (1024, _PREFIXES_BIN)
    value, prefix = _find_prefix(value, base, prefixes)

    precision = '.3g' if -99.95 < value < 99.95 else '.0f'
    prefix = prefix.strip()
    unit = prefix if si else f'{prefix}iB' if prefix else 'B'
    return f'{value:{precision}}{unit}'


class _Si(ObjectProxy):
    __slots__ = '_self_si',

    def __init__(self, value: float | int = 0, si: bool = True):
        super().__init__(value)
        self._self_si = si

    def __str__(self) -> str:
        return _num_repr(self.__wrapped__, self._self_si)

    def __repr__(self):
        return f'{type(self).__name__}({self})'

    def __format__(self, format_spec: str) -> str:
        return str(self).__format__(format_spec)

    def __reduce_ex__(self, _) -> tuple:  # Else no serialization
        return type(self), (self.__wrapped__, self._self_si)


def si(value: _IntOrFloat) -> _IntOrFloat:
    """Mix value with human-readable formatting,
    uses metric prefixes. Returns exact subtype of value.

    >>> s = si(10_000)
    >>> print(s)
    10k
    """
    return cast(_IntOrFloat, _Si(value, si=True))


def si_bin(value: _IntOrFloat) -> _IntOrFloat:
    """Treats value as size in bytes, mixes it with binary prefix.
    Returns exact subtype of value.

    >>> s = si_bin(4096)
    >>> print(s, s + 5)
    4KiB 4101

    .. _Human readable bytes count
       https://programming.guide/java/formatting-byte-size-to-human-readable-format.html
    """
    return cast(_IntOrFloat, _Si(value, si=False))



from __future__ import annotations

__all__ = ['memoize']

import argparse
import asyncio
import concurrent.futures as cf
import enum
import functools
from collections import Counter
from collections.abc import (Callable, Hashable, Iterable, KeysView,
                             MutableMapping)
from contextlib import ExitStack
from dataclasses import dataclass, field
from threading import RLock
from typing import (Any, ClassVar, Generic, Literal, NamedTuple, SupportsInt,
                    TypeVar, cast)
from weakref import WeakValueDictionary

from .._repr import si_bin
from .._sizeof import sizeof
from .reusable import make_loop
from .util import make_key


class _Empty(enum.Enum):
    token = 0


_T = TypeVar('_T')
_F = TypeVar('_F', bound=Callable)
_BatchedFn = TypeVar('_BatchedFn', bound=Callable[[list], Iterable])
_Policy = Literal['raw', 'lru', 'mru']
_KeyFn = Callable[..., Hashable]
_empty = _Empty.token


def _unit_size(x):
    return 1


@dataclass(repr=False)
class _Node(Generic[_T]):
    __slots__ = ('value', 'size')
    value: _T
    size: int

    def __repr__(self) -> str:
        return repr(self.value)


class Stats(argparse.Namespace):
    def __init__(self, **kwargs):
        self.__dict__ = Counter()
        self.__dict__.update(**kwargs)

    def __getattr__(self, name: str) -> int:
        return self.__dict__[name]


class _IStore(Generic[_T]):
    def __len__(self) -> int:
        raise NotImplementedError

    def store_clear(self) -> None:
        raise NotImplementedError

    def store_get(self, key: Hashable) -> _Node[_T] | None:
        raise NotImplementedError

    def store_set(self, key: Hashable, node: _Node[_T]) -> None:
        raise NotImplementedError

    def can_shrink_for(self, size: int) -> bool:
        raise NotImplementedError


@dataclass(repr=False)
class _InitializedStore:
    capacity: int
    size_fn: Callable[[object], int]
    size: int = field(default=0, init=False)
    stats: Stats = field(default_factory=Stats, init=False)


@dataclass(repr=False)
class _DictMixin(_InitializedStore, _IStore[_T]):
    lock: RLock = field(default_factory=RLock, init=False)

    def clear(self):
        with self.lock:
            self.store_clear()
            self.size = 0

    def keys(self) -> KeysView:
        raise NotImplementedError

    def __getitem__(self, key: Hashable) -> _T | _Empty:
        with self.lock:
            if node := self.store_get(key):
                self.stats.hits += 1
                return node.value
        return _empty

    def __setitem__(self, key: Hashable, value: _T) -> None:
        with self.lock:
            self.stats.misses += 1
            size = int(self.size_fn(value))
            if (self.capacity < 0  # is unbound
                    or self.size + size <= self.capacity  # has free place
                    or (size < self.capacity and self.can_shrink_for(size))):
                self.store_set(key, _Node(value, size))
                self.size += size


@dataclass(repr=False)
class _ReprMixin(_InitializedStore, _IStore[_T]):
    refs: ClassVar[MutableMapping[int, _ReprMixin]] = WeakValueDictionary()

    def __post_init__(self) -> None:
        self.refs[id(self)] = self

    def __repr__(self) -> str:
        args = [
            f'items={len(self)}',
            f'used={si_bin(self.size)}',
            f'total={si_bin(self.capacity)}',
        ]
        if any(vars(self.stats).values()):
            args.append(f'stats={self.stats}')
        return f'{type(self).__name__}({", ".join(args)})'

    @classmethod
    def status(cls) -> str:
        return '\n'.join(
            f'{id_:x}: {value!r}' for id_, value in sorted(cls.refs.items()))


@dataclass(repr=False)
class _Store(_ReprMixin[_T], _DictMixin[_T]):
    store: dict[Hashable, _Node[_T]] = field(default_factory=dict)

    def __len__(self) -> int:
        return len(self.store)

    def keys(self) -> KeysView:
        return self.store.keys()

    def store_clear(self) -> None:
        self.store.clear()

    def store_get(self, key: Hashable) -> _Node[_T] | None:
        return self.store.get(key)

    def store_set(self, key: Hashable, node: _Node[_T]) -> None:
        self.store[key] = node


class _HeapCache(_Store[_T]):
    def can_shrink_for(self, size: int) -> bool:
        return False


class _LruCache(_Store[_T]):
    drop_recent = False

    def store_get(self, key: Hashable) -> _Node[_T] | None:
        if node := self.store.pop(key, None):
            self.store[key] = node
            return node
        return None

    def can_shrink_for(self, size: int) -> bool:
        size = 0
        while self.store and self.size + size > self.capacity:
            if self.drop_recent:
                self.size -= self.store.popitem()[1].size
            else:
                self.size -= self.store.pop(next(iter(self.store))).size
            self.stats.dropped += 1
        return True


class _MruCache(_LruCache[_T]):
    drop_recent = True


# -------------------------------- wrapping --------------------------------


def _memoize(cache: _DictMixin, key_fn: _KeyFn, fn: _F) -> _F:
    def wrapper(*args, **kwargs):
        key = key_fn(*args, **kwargs)

        if (value := cache[key]) is not _empty:
            return value

        # NOTE: fn() is not within lock
        cache[key] = value = fn(*args, **kwargs)
        return value

    wrapper.cache = cache  # type: ignore[attr-defined]
    return cast(_F, functools.update_wrapper(wrapper, fn))


# ----------------------- wrapper with batching support ----------------------


class _Job(NamedTuple):
    token: Any
    future: asyncio.Future | cf.Future


def _dispatch(
    fn: Callable[[list], Iterable],
    evict: Callable[[Hashable], object],
    queue: dict[Hashable, _Job],
):
    jobs = {**queue}
    queue.clear()

    try:
        *values, = fn([job.token for job in jobs.values()])

        # py3.10: zip(..., strict=True)
        if len(values) != len(jobs):
            raise RuntimeError(  # noqa: TRY301
                'Input batch size is not equal to output: '
                f'{len(values)} != {len(jobs)}')

        for job, value in zip(jobs.values(), values):
            job.future.set_result(value)

    except BaseException as exc:  # noqa: BLE001
        for key, job in jobs.items():
            evict(key)
            job.future.set_exception(exc)


def _memoize_batched_aio(key_fn: _KeyFn, fn: _BatchedFn) -> _BatchedFn:
    assert callable(fn)
    futs: dict[Hashable, asyncio.Future] = {}
    queue: dict[Hashable, _Job] = {}
    loop = make_loop()

    def _load(token) -> asyncio.Future:
        key = key_fn(token)
        if result := futs.get(key):
            return result

        loop = asyncio.get_running_loop()
        futs[key] = future = loop.create_future()
        queue[key] = _Job(token, future)
        if len(queue) == 1:
            loop.call_soon(_dispatch, fn, futs.pop, queue)

        return future

    async def _load_many(tokens: Iterable) -> tuple:
        return await asyncio.gather(*map(_load, tokens))

    def wrapper(tokens: Iterable) -> tuple:
        coro = _load_many(tokens)
        return asyncio.run_coroutine_threadsafe(coro, loop).result()

    wrapper.stage = futs  # type: ignore[attr-defined]
    return cast(_BatchedFn, functools.update_wrapper(wrapper, fn))


def _memoize_batched(cache: _DictMixin, key_fn: _KeyFn,
                     fn: _BatchedFn) -> _BatchedFn:
    assert callable(fn)
    lock = RLock()
    futs = WeakValueDictionary[Hashable, cf.Future]()
    queue: dict[Hashable, _Job] = {}

    def _load(stack: ExitStack, key: Hashable, token: object) -> cf.Future:
        with lock:
            if result := futs.get(key):
                return result

            futs[key] = future = cf.Future()  # type: ignore[var-annotated]
            queue[key] = _Job(token, future)
            if len(queue) == 1:
                stack.callback(_dispatch, fn, futs.pop, queue)

        return future

    def wrapper(tokens: Iterable) -> list:
        keyed_tokens = [(key_fn(t), t) for t in tokens]

        # Try to hit
        misses = {}
        hits = {}
        with cache.lock:
            for k, t in dict(keyed_tokens).items():
                if (r := cache[k]) is not _empty:
                    hits[k] = r
                else:
                    misses[k] = t

        futs: dict[Hashable, cf.Future] = {}
        with ExitStack() as stack:
            futs |= {k: _load(stack, k, t) for k, t in misses.items()}
        cf.wait(futs.values(), return_when='FIRST_EXCEPTION')

        # Process misses
        with cache.lock:
            for k, f in futs.items():
                hits[k] = cache[k] = f.result()
        return [hits[k] for k, _ in keyed_tokens]

    wrapper.cache = cache  # type: ignore[attr-defined]
    wrapper.stage = futs  # type: ignore[attr-defined]
    return cast(_BatchedFn, functools.update_wrapper(wrapper, fn))


# ----------------------------- factory wrappers -----------------------------


def memoize(
    capacity: SupportsInt | None,
    *,
    batched: bool = False,
    policy: _Policy = 'raw',
    key_fn: _KeyFn = make_key,
    bytesize: bool = True,
) -> Callable[[_F], _F] | Callable[[_BatchedFn], _BatchedFn]:
    """Returns dict-cache decorator.

    Parameters:
    - capacity - max size in bytes if `bytesize` is set, otherwise objects.
      Cache is unbound if None set.
    - policy - eviction policy, either "raw" (no eviction), or "lru"
      (evict oldest), or "mru" (evict most recent).
    - bytesize - if set limits bytes, not objects.
    """
    capacity = int(capacity) if capacity is not None else -1
    if capacity == 0:
        return lambda fn: fn

    caches: dict[str, type[_Store]] = {
        'raw': _HeapCache,
        'lru': _LruCache,
        'mru': _MruCache,
    }
    size_fn = sizeof if bytesize else _unit_size

    if (cache_cls := caches.get(policy)) is not None:
        if capacity < 0:
            cache_cls = _HeapCache
        cache = cache_cls(capacity, size_fn)
        if batched:
            return functools.partial(_memoize_batched, cache, key_fn)
        return functools.partial(_memoize, cache, key_fn)
    raise ValueError(f'Unknown policy: "{policy}". '
                     f'Only "{set(caches)}" are available')



from __future__ import annotations

import torch
from matplotlib import pyplot as plt
from torch import nn
from tqdm.auto import tqdm

import glow.nn
from glow import metrics as m

metrics: tuple[m.Metric, ...] = (
    m.Lambda(m.accuracy_),
    m.Confusion(
        acc=m.accuracy,
        accb=m.accuracy_balanced,
        iou=m.iou,
        kappa=m.kappa,
    ),
)

c = 8
b = 128
true = torch.randint(c, size=[b])
pred = torch.randn(b, c, requires_grad=True)


class Model(nn.Module):
    def forward(self, x):
        return net.param


net = Model()
net.param = nn.Parameter(data=pred, requires_grad=True)

optim = glow.nn.RAdam(net.parameters())
cm_grad = m.ConfusionGrad()

plt.ion()
_, ax = plt.subplots(ncols=4)
ax[2].plot(true.numpy())
with tqdm(range(32)) as pbar:
    for _ in pbar:
        for _ in range(64):
            net.zero_grad()
            cm = cm_grad(net(None), true)
            loss = -m.accuracy(cm)
            # loss = -m.accuracy_balanced(cm)
            # loss = -m.kappa(cm)
            loss.backward()
            optim.step()
            pbar.set_postfix({'score': -loss.detach_().item()})
            cm.detach_()

        ax[0].imshow(pred.detach().numpy())
        ax[1].imshow(pred.detach().softmax(1).numpy())
        ax[2].cla()
        ax[2].plot(sorted(zip(true.numpy(), pred.detach().argmax(1).numpy())))
        ax[3].imshow(cm.numpy(), vmax=1 / c)

        plt.pause(1e-2)

with torch.no_grad():
    meter = m.compose(*metrics)
    d = meter.send((pred, true))
    print(', '.join(f'{k}: {v:.3f}' for k, v in d.scalars.items()))
    print(', '.join(f'{k}: {v.item():.3f}'
                    for k, v in d.tensors.items() if v.numel() == 1))



from __future__ import annotations

__all__ = ['ConvCtx']

import warnings
from dataclasses import dataclass, replace
from typing import Literal

from torch import nn

from .lazy import LazyBlurPool2d
from .simple import Decimate2d
from .util import ActivationFn, LazyConvFn, LazyNormFn


@dataclass(frozen=True)
class ConvCtx:
    """
    Parity:
    - 0: kernel = 2 * overlap + stride
    - 1: kernel = 2 * overlap + 1
    """
    conv_fn: LazyConvFn = nn.LazyConv2d
    # TODO: accept kernel_pool
    # TODO: compute padding from kernel_pool
    parity: Literal[0, 1] = 1
    overlap: int = 1
    pad: bool = True

    norm: LazyNormFn = nn.LazyBatchNorm2d

    activation: ActivationFn = nn.ReLU
    inplace: bool = True

    def _get_k(self, stride: int, overlap: int | None = None) -> int:
        if overlap is None:
            overlap = self.overlap
        return (stride if self.parity == 0 else 1) + 2 * overlap

    def _get_p(self, kernel: int, stride: int, dilation: int) -> int:
        assert stride == 1 or dilation == 1, \
            'one of stride/dilation should be always 1'
        if stride == 1 and kernel % 2 == 0 and dilation % 2 != 0:
            raise ValueError('Even kernel with odd dilation is not supported')

        if self.parity == 0:
            total_padding = kernel - stride
            assert total_padding >= 0, \
                'kernel should be same or greater than stride'
            assert total_padding >= 0
        else:
            total_padding = kernel - 1

        assert total_padding % 2 == 0, \
            'padding is not symmetric, offset kernel by 1'
        if not self.pad:
            return 0
        return (total_padding // 2) * dilation

    def _invert(self) -> ConvCtx:
        return replace(self, parity=1 - self.parity)

    def conv(self,
             dim: int,
             kernel: int | None = None,
             stride: int = 1,
             overlap: int | None = None,
             dilation: int = 1,
             groups: int | None = 1,
             bias: bool = True) -> nn.modules.conv._ConvNd:
        """
        Basic convolutions, pass kernel explicitly:
        ```
        >>> ctx.conv(64, 1)  # conv1x1, 64f
        >>> ctx.conv(64, 3)  # conv3x3/pad=1, 64f
        >>> ctx.conv(64, 3, groups=64)  # dw-conv3x3/pad=1
        ```
        Dilations, will fail with stride != 1:
        ```
        >>> ctx.conv(64, 3, dilation=2)  # conv3x3/d2/pad=2
        ```

        Strided convolutions.
        Better use implicit kernel if parity switch is desired.
        Even mode (parity = 0):
        ```
        >>> ctx.conv(64, stride=2)     # conv4x4/s2/pad=1
        >>> ctx.conv(64, 4, stride=2)  # same, explicit
        >>> ctx.conv(64, 3, stride=2)  # fail
        ```
        Odd mode (parity = 1), can introduce aliasing:
        ```
        >>> ctx.conv(64, stride=2)     # conv3x3/s2/pad=1
        >>> ctx.conv(64, 3, stride=2)  # same, explicit
        >>> ctx.conv(64, 4, stride=2)  # fail
        ```
        """
        if kernel is None:
            kernel = self._get_k(stride, overlap)

        elif stride == 1 and kernel % 2 == 0:
            warnings.warn(
                f'Parity is altered because of {kernel=}', stacklevel=2)

        elif stride != 1 and kernel != self.parity:
            raise ValueError(f'Used kernel does not match used parity '
                             f'({self.parity=}). Network is likely to break')

        padding = self._get_p(kernel, stride, dilation)
        groups = groups or dim
        return self.conv_fn(dim, kernel, stride, padding, dilation, groups,
                            bias)

    def avg_pool(self,
                 stride: int = 2,
                 overlap: int | None = None) -> Decimate2d | nn.AvgPool2d:
        assert stride > 1

        kernel = self._get_k(stride, overlap)
        if kernel == 1:
            return Decimate2d(stride)

        padding = self._get_p(kernel, stride, 1)
        return nn.AvgPool2d(kernel, stride, padding)

    def max_pool(self,
                 stride: int = 2,
                 overlap: int | None = None) -> Decimate2d | nn.MaxPool2d:
        assert stride > 1

        kernel = self._get_k(stride, overlap)
        if kernel == 1:
            return Decimate2d(stride)

        padding = self._get_p(kernel, stride, 1)
        return nn.MaxPool2d(kernel, stride, padding)

    def conv_unpool(self,
                    dim: int,
                    stride: int = 2,
                    overlap: int | None = None,
                    groups: int | None = 1,
                    bias: bool = True) -> nn.ConvTranspose2d:
        groups = groups or dim
        kernel = self._get_k(stride, overlap)
        padding = self._get_p(kernel, stride, 1)
        return nn.LazyConvTranspose2d(dim, kernel, stride, padding, 0, groups,
                                      bias)

    def activation_(self) -> nn.Module:
        if self.inplace:
            return self.activation(inplace=True)
        return self.activation()

    def norm_act(self, act: bool = True) -> list[nn.Module]:
        norm = self.norm()
        return [norm, self.activation_()] if act else [norm]

    def conv_norm(self, mod: nn.modules.conv._ConvNd) -> nn.Sequential:
        assert self.norm
        mod.bias = None
        return nn.Sequential(mod, self.norm())

    def conv_norm_act(self, mod: nn.modules.conv._ConvNd) -> nn.Sequential:
        norms: list[nn.Module] = []
        if self.norm:
            mod.bias = None  # Normalization overrides this with its own bias
            norms += [self.norm()]
        return nn.Sequential(mod, *norms, self.activation_())

    def norm_act_conv(self, mod: nn.modules.conv._ConvNd) -> nn.Sequential:
        assert self.norm
        mod.bias = None
        return nn.Sequential(self.norm(), self.activation_(), mod)

    def blur_pool(self,
                  stride: int = 2,
                  overlap: int | None = None) -> nn.Module:
        """Replacement (and generalization) for AvgPool2d(2, 2)"""
        kernel = self._get_k(stride, overlap)
        padding = self._get_p(kernel, stride, 1)
        return LazyBlurPool2d(kernel, stride, padding)

    def max_blur_pool(self, stride: int = 2, overlap: int = 1) -> nn.Module:
        """
        Replacement for MaxPool2d(2, 2)
        Only for even inputs. When applied to odd inputs, loses 1 sample.
        """
        assert self.parity == 0
        return nn.Sequential(
            # switches parity
            nn.MaxPool2d(2, 1),
            # correct pool
            self._invert().blur_pool(stride, overlap),
        )

    def conv_blur_pool(self,
                       dim: int,
                       stride: int = 2,
                       overlap: int = 1) -> nn.Module:
        """Replacement for [Conv2d(..., 3, 2, 1), norm, act]"""
        return nn.Sequential(
            self.conv(dim, 3),
            self.norm(),
            self.activation_(),
            self.blur_pool(stride, overlap),
        )



from __future__ import annotations

__all__ = ['Uid']

import math
import re
import string
from functools import lru_cache
from typing import SupportsInt
from uuid import UUID, uuid4

ALPHABET = string.digits + string.ascii_letters
ALPHABET = ''.join(sorted({*ALPHABET} - {*'0O1Il'}))

_BASE = len(ALPHABET)  # 57
_LEN = math.ceil(128 / math.log2(_BASE))  # 22

_TABLE = ALPHABET.encode('ascii').ljust(256, b'\0')
_NUMBERS = {s: i for i, s in enumerate(ALPHABET)}
_REGEX = re.compile(f'^[{ALPHABET}]{{{_LEN}}}$')


@lru_cache  # Small performance optimization
def base57_encode(number: int) -> str:
    out = bytearray(_LEN)
    for i in range(_LEN - 1, -1, -1):
        number, out[i] = divmod(number, _BASE)
    return out.translate(_TABLE).decode('ascii')


@lru_cache  # Small performance optimization
def base57_decode(shortuuid: str) -> int:
    if not _REGEX.fullmatch(shortuuid):
        raise ValueError('invalid shortuuid format')
    out = 0
    for char in shortuuid:
        out = out * _BASE + _NUMBERS[char]
    return out


class Uid(UUID):
    """Subclass of UUID with support of short-uuid serialization format.

    Uses base57 instead of hex for serialization.

    base57 uses lowercase and uppercase letters and digits,
    excluding similar-looking characters such as l, 1, I, O and 0,
    and it doesn't use URL-unsafe +, /, = characters (opposed to base64).

    UUIDs encoded with base57 have length of 22 characters, while
    with hex (default) - 32 characters.

    Uid can be created directly from UUID:

        >>> u = UUID('3b1f8b40-222c-4a6e-b77e-779d5a94e21c')
        >>> Uid(u)
        Uid('CXc85b4rqinB7s5J52TRYb')
        >>> str(Uid(u))
        'CXc85b4rqinB7s5J52TRYb'

    Or from string representation of short-uuid:

        >>> Uid('CXc85b4rqinB7s5J52TRYb')
        Uid('CXc85b4rqinB7s5J52TRYb')

    Simplified and more optimized (2-3x faster on average) fork of
    [shortuuid](https://github.com/skorokithakis/shortuuid)
    """
    def __init__(self, obj: str | SupportsInt):
        """
        Creates Uid either from str (parsing it as short-uuid) or
        from int()-compatible type
        """
        if not isinstance(obj, (str, SupportsInt)):
            raise TypeError('Either int, string or UUID required. '
                            f'Got {type(obj)}')

        value = base57_decode(obj) if isinstance(obj, str) else int(obj)
        super().__init__(int=value)

    def __str__(self) -> str:
        return base57_encode(int(self))

    @classmethod
    def __get_validators__(cls):  # Required for Pydantic
        yield cls

    @classmethod
    def __modify_schema__(cls, field_schema: dict):  # Required for OpenAPI
        field_schema.update(
            examples=[str(cls.v4()) for _ in range(2)],
            type='string',
            format=None,
            pattern=_REGEX.pattern,
        )

    @classmethod
    def v4(cls) -> Uid:
        """Alias for Uid(uuid.uuid4())"""
        return cls(uuid4())



__all__ = ['env']

from collections import ChainMap
from collections.abc import Iterator
from contextlib import contextmanager

from ..core import repr_as_obj


class _Env(ChainMap):
    """Environment with scopes.
    The outermost scope takes highest priority.

    As example:

        >>> print(env)  # Empty in global scope
        _Env()
        >>> with env(a=1):  # Overrides `a` from inner scope
        ...     print(env)
        ...     with env(a=2, b=3):  # Set defaults for `a` and `b`
        ...         print(env)
        ...     print(env)
        _Env(a=1)
        _Env(a=1, b=3)
        _Env(a=1)
        >>> print(env)  # Reset to initial state
        _Env()

    """
    @contextmanager
    def __call__(self, **items) -> Iterator[None]:
        self.maps.append(items)
        try:
            yield
        finally:
            self.maps.pop()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr_as_obj({**self})})'


env = _Env()
